/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LSTDPP_BASEREADER_H
#define LSTDPP_BASEREADER_H

#include <stdint.h>
#include <fstream>
#include <vector>
#include <string>
#include <sstream>

#include "CrateBoard.h"

namespace lstdpp128 {

class BaseReader {

public:
	/**
	 * Constructor.
	 */
	BaseReader();

	/**
	 * Destructor.
	 */
	virtual ~BaseReader();

	/**
	 * Returns the memory block.
	 */
	int32_t * buffer() const;

	/**
	 * Returns the version.
	 */
	uint32_t version() const;

	/**
	 * Returns the time base.
	 */
	uint32_t timeBase() const;

	/**
	 * Returns the total number of boards.
	 */
	uint32_t totalNumberOfBoards() const;

	/**
	 * Opens the file. Returns false if problem opening the file.
	 */
	virtual bool open(std::string const & fileName) = 0;

	/**
	 * Returns the number of int32_t elements read.
	 */
	virtual int32_t read() = 0;

protected:
	/**
	 * Initialises the reader. The reader allocates a block of size int32_t elements.
	 * The size must be a multiple of EVENT_SIZE_32.
	 */
	void init(int32_t size);

	typedef std::ifstream::pos_type FileSize;

	int32_t _size;
	int32_t _sizeInBytes;

	int32_t * _buffer;
	std::ifstream _dataFile;

	FileSize _fileSize;
	uint32_t _version;
	uint32_t _timeBase;
	uint32_t _totalNumberOfBoards;
	int32_t * _boardHeader;
};

/**
 * Loads the list of values.
 */
template<typename Type>
static std::vector<Type> loadList(std::string const & fileName) {

	std::vector<Type> result;

	std::ifstream file(fileName.c_str());
	std::string line;

	while (getline(file, line)) {
		std::istringstream iv(line);
		int16_t value;
		iv >> value;
		result.push_back(value);
	}

	return result;
}

/**
 * Prints the list of values.
 */
template<typename Type>
static std::string printList(std::vector<Type> const & values) {

	std::ostringstream v;
	v << "[";

	for (int i = 0; i < values.size() - 1; ++i) {
		v << values[i] << ", ";
	}
	if (values.size() > 0) {
		v << values[values.size() - 1];
	}
	v << "]";

	return v.str();
}

}

#endif
