/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CrateBoard.h"

#include <map>

using namespace std;

namespace lstdpp128 {

ListModeContext listModeContext;

std::ostream& operator<< (std::ostream& os, ListModeContext const & header) {

	os << "version   = " << header.version << endl;
	os << "time base = " << header.timeBase << endl;
	os << "#events   = " << header.nbEvents << endl;
	os << "#crates   = " << header.crateBoard.nbCrates << endl << endl;

	for (int c = 0; c < header.crateBoard.nbCrates; c++) {
		os << "crate " << c << " #boards = " << header.crateBoard.crates[c].nbBoards << endl << endl;
		for (int b = 0; b < header.crateBoard.crates[c].nbBoards; b++) {
			os << "\t" << "board " << b << endl;
			os << "\ttype            = " << (int)header.crateBoard.crates[c].boards[b].boardType << endl;
			os << "\ttype name       = " << BoardTypeName[header.crateBoard.crates[c].boards[b].boardType] << endl;
			os << "\tevent type      = " << header.crateBoard.crates[c].boards[b].eventType << endl;
			os << "\ttime resolution = " << BoardTimeResolution[header.crateBoard.crates[c].boards[b].boardType] << " nanoseconds" << endl;
			os << "\t#channels       = " << header.crateBoard.crates[c].boards[b].nbChannels << endl << endl;
		}
	}

	return os;
}

std::string writeListModeContextToBinary(const ListModeContext & listModeContext) {

	// Count the number of boards.
	int boardSize = 0;

	for (int c = 0; c < listModeContext.crateBoard.nbCrates; c++) {
		for (int b = 0; b < listModeContext.crateBoard.crates[c].nbBoards; b++) {
			++boardSize;
		}
	}

	// Allocate the data.
	uint32_t buffer[4 + boardSize];
	buffer[0] = listModeContext.version;
	buffer[1] = listModeContext.timeBase;
	buffer[2] = listModeContext.nbEvents;
	buffer[3] = boardSize;

	int i = 0;
	for (int c = 0; c < listModeContext.crateBoard.nbCrates; c++) {
		for (int b = 0; b < listModeContext.crateBoard.crates[c].nbBoards; b++) {

			int32_t eventType = listModeContext.crateBoard.crates[c].boards[b].eventType;
			int32_t crate = listModeContext.crateBoard.crates[c].boards[b].crate;
			int32_t nbChannels = listModeContext.crateBoard.crates[c].boards[b].nbChannels;
			int32_t boardType = listModeContext.crateBoard.crates[c].boards[b].boardType;

			buffer[4 + i] = boardType & 0x3F;
			buffer[4 + i] |= (nbChannels & 0x3F) << 6;
			buffer[4 + i] |= (crate & 0xF) << 12;
			buffer[4 + i] |= (eventType & 0xFFFF) << 16;

			++i;
		}
	}

	// Return the binary content.
	return string(reinterpret_cast<const char *>(buffer), (4 + boardSize) * sizeof(uint32_t));
}

uint32_t readListModeContextFromBinary(ListModeContext & listModeContext, uint32_t version, uint32_t timeBase, uint32_t nbEvents, uint32_t nbBoards, uint32_t* data) {

	listModeContext.version = version;
	listModeContext.timeBase = timeBase;
	listModeContext.nbEvents = nbEvents;
	listModeContext.crateBoard.nbCrates = 0;

	// Determine the number of crates and the number of boards per crate.
	map<uint8_t, uint16_t> boardSizeMap;

	for (uint32_t i = 0; i < nbBoards; ++i) {

		uint8_t crate = (uint8_t)((data[i] >> 12) & 0xF);
		boardSizeMap[crate]++;

		if (crate >= listModeContext.crateBoard.nbCrates) {
			listModeContext.crateBoard.nbCrates = crate + 1;
		}
	}

	// Allocate the crates.
	listModeContext.crateBoard.crates = new Crate[listModeContext.crateBoard.nbCrates];

	// Allocate the boards.
	for (uint32_t c = 0; c < listModeContext.crateBoard.nbCrates; c++) {

		uint16_t crateNbBoards = boardSizeMap[c];
		listModeContext.crateBoard.crates[c].nbBoards = crateNbBoards;
		listModeContext.crateBoard.crates[c].boards = new Board[crateNbBoards];
	}

	// Fill the crates and boards.
	map<uint8_t, uint16_t> boardIndexMap;

	for (uint32_t i = 0; i < nbBoards; ++i) {

		uint8_t crate = (uint8_t)((data[i] >> 12) & 0xF);
		uint16_t board = boardIndexMap[crate];

		listModeContext.crateBoard.crates[crate].boards[board].eventType = (uint16_t)((data[i] >> 16) & 0xFFFF);
		listModeContext.crateBoard.crates[crate].boards[board].crate = crate;
		listModeContext.crateBoard.crates[crate].boards[board].nbChannels = (uint16_t)((data[i] >> 6) & 0x3F);
		listModeContext.crateBoard.crates[crate].boards[board].boardType = (BoardType) (data[i] & 0x3F);

		boardIndexMap[crate]++;
	}

	return nbBoards;
}

void releaseListModeContext(ListModeContext & listModeContext) {

	for (int c = 0; c < listModeContext.crateBoard.nbCrates; c++) {
		delete [] listModeContext.crateBoard.crates[c].boards;
	}

	delete [] listModeContext.crateBoard.crates;
}

BoardIterator::BoardIterator() :
	_currentCrate(0),
	_currentBoard(0) {
}

uint8_t BoardIterator::crate() const {
	return _currentCrate;
}

uint16_t BoardIterator::board() const {
	return _currentBoard;
}

bool BoardIterator::end() const {

	if (_currentCrate == listModeContext.crateBoard.nbCrates) {
		return true;
	}
	return false;
}

Board & BoardIterator::operator* () const {
	return listModeContext.crateBoard.crates[_currentCrate].boards[_currentBoard];
}

bool BoardIterator::operator== (BoardIterator const & i) const {
	return (_currentCrate == i._currentCrate && _currentBoard == i._currentBoard);
}

bool BoardIterator::operator!= (BoardIterator const & i) const {
	return (_currentCrate != i._currentCrate || _currentBoard != i._currentBoard);
}

BoardIterator & BoardIterator::operator++ () {

	_currentBoard++;
	if (_currentBoard == listModeContext.crateBoard.crates[_currentCrate].nbBoards) {
		_currentCrate++;
		_currentBoard = 0;
	}

	return *this;
}

}
