/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Event.h"

#include <iomanip>
#include <cmath>
#include <string.h>
#include <sstream>

using namespace std;

namespace lstdpp128 {

time_type Event::time() const {
	return (((int64_t)rollover << 32) | timestamp);
}

void Event::setTime(time_type value) {

	timestamp = (value & 0xFFFFFFFF);
	rollover = ((value >> 32) & 0xFFFF);
}

std::string eventColumnsToString(BoardType type) {

	ostringstream os;

	os 	<< setw(10) << (int)type
			<< setw(10) << "time"
			<< setw(8) << "roll."
			<< setw(12) << "timestamp"
			<< setw(8) << "crate"
			<< setw(8) << "board"
			<< setw(8) << "channel"
			<< setw(11) << "board type";

	return os.str();
}

std::ostream& operator<< (std::ostream& os, Event const & e) {

	return os << dec
			<< setw(20) << e.time()
			<< setw(8) << e.rollover
			<< setw(12) << e.timestamp
			<< setw(8) << (int)e.crate
			<< setw(8) << (int)e.board
			<< setw(8) << e.channel
			<< setw(11) << (int)e.boardType();
}

std::string printBinary(int32_t * buffer) {

	ostringstream os;

	os << hex << setw(8) << setfill('0') << (uint32_t)(*(buffer))
											<< " " << setw(8) << setfill('0') << (uint32_t)(*(buffer + 1))
											<< " " << setw(8) << setfill('0') << (uint32_t)(*(buffer + 2))
											<< " " << setw(8) << setfill('0') << (uint32_t)(*(buffer + 3))
											<< setw(0) << setfill(' ') << dec;

	return os.str();
}

int32_t EventBlock::maxId = 0;

EventBlock * EventBlock::split() {

	time_type midTime = (minTime() + maxTime()) / 2;

	return split(midTime);
}

EventBlock * EventBlock::split(time_type midTime) {

	EventBlock * upperBlock = new EventBlock();

	Event event;
	event.setTime(midTime);

	multiset<Event, CompareEvents>::const_iterator e = events.upper_bound(event);

	upperBlock->events = multiset<Event, CompareEvents>(e, events.end());
	events = multiset<Event, CompareEvents>(events.begin(), e);

	return upperBlock;
}

std::ostream& operator<< (std::ostream& os, EventBlock const & b) {
	if (b.events.empty()) {
		return os << b.id << " : empty";
	}
	return os << b.id << " : " << b.minTime() << " - " << b.maxTime() << " (" << b.events.size() << ")";
}

void readEventV1(Event & event, int32_t * buffer) {

	BoardType boardType = event.boardType();

	// The number of bits for timestamp.
	int numberOfBitsOfTimestamp = 32;

	// In case of 1724 PHA, the number of bits for timestamp is 30.
	if (boardType == C1724_PHA) {
		numberOfBitsOfTimestamp = 30;
		event.timestamp = (uint32_t)(*(buffer + 1) & 0x3FFFFFFF);
	}
	else if (boardType == C1725_PHA) {
		numberOfBitsOfTimestamp = 31;
		event.timestamp = (uint32_t)(*(buffer + 1) & 0x7FFFFFFF);
	}
	else {
		event.timestamp = (uint32_t)(*(buffer + 1) & 0xFFFFFFFF);
	}

	// Copy data.
	memcpy(reinterpret_cast<void *>(event.data), reinterpret_cast<const void *>(buffer + 2), 8);

	// Use printBinary(buffer) to debug if necessary.

	// Calculate the absolute time.
	time_type absoluteTime = (((int64_t)event.rollover << numberOfBitsOfTimestamp) | event.timestamp);

	// Testing if we must apply the time resolution or not in case of a board that has relative time.
	if (BoardRelativeTime[boardType]) {
		// Set the absolute time. Here the event is still in relative time.
		absoluteTime = absoluteTime * event.timeResolution();

		// We can set the absolute time that changes the rollover and timestamp.
		event.setTime(absoluteTime);

		if (absoluteTime != event.time()) {
			cerr << "Problem while converting relative time to absolute time" << endl;
		}
	}
	else {
		// We can set the absolute time that changes the rollover and timestamp.
		event.setTime(absoluteTime);
	}
}

void readEventV2(Event & event, int32_t * buffer) {

	BoardType boardType = event.boardType();

	// The number of bits for timestamp.
	int numberOfBitsOfTimestamp = 32;

	event.timestamp = (uint32_t)(*(buffer + 1) & 0xFFFFFFFF);

	// Copy data.
	memcpy(reinterpret_cast<void *>(event.data), reinterpret_cast<const void *>(buffer + 2), 8);

	// Use printBinary(buffer) to debug if necessary.

	// Calculate the absolute time.
	time_type absoluteTime = (((int64_t)event.rollover << numberOfBitsOfTimestamp) | event.timestamp);

	// We can set the absolute time that changes the rollover and timestamp.
	event.setTime(absoluteTime);
}

bool readEvent(Event & event, int32_t * buffer) {

	// Read crate, board, channel.
	uint8_t crate = (uint8_t)((*buffer >> 28) & 0xF);
	uint8_t board = (uint8_t)((*buffer >> 22) & 0x3F);

	if (crate >= listModeContext.crateBoard.nbCrates) {
		// No output because the log file can be very large.
		return false;
	}

	if (board >= listModeContext.crateBoard.crates[event.crate].nbBoards) {
		// No output because the log file can be very large.
		return false;
	}

	event.crate = crate;
	event.board = board;
	event.channel = (uint8_t)((*buffer >> 16) & 0x3F);
	event.rollover = (uint16_t)(*buffer & 0xFFFF);

	// Check the lst version to read the event.
	if (listModeContext.version == 1) {
		readEventV1(event, buffer);
	}
	else {
		readEventV2(event, buffer);
	}

	return true;
}

void writeEventV1(Event const & event, int32_t * buffer) {

	Event relativeTimeEvent(event);

	// Testing if we must apply the time resolution or not in case of a board that has relative time.
	if (BoardRelativeTime[event.boardType()]) {

		// We reset the event time to the relative time because we do not want to change the header of the lsto.
		time_type relativeTime = event.time() / event.timeResolution();

		// We can set the relative time that changes the rollover and timestamp.
		relativeTimeEvent.setTime(relativeTime);
	}

	*buffer = 0;
	*buffer |= ((relativeTimeEvent.crate & 0xF) << 28);
	*buffer |= ((relativeTimeEvent.board & 0x3F) << 22);
	*buffer |= ((relativeTimeEvent.channel & 0x3F) << 16);

	BoardType boardType = relativeTimeEvent.boardType();

	if (boardType == C1724_PHA) {

		// The card timestamp is on 30bits with 2 bits of zero.
		// The rollover is calculated by taking the 2 bits of the actual timestamp (>> 30 for 32 bits = 2 bits).
		int32_t rollover = ((relativeTimeEvent.rollover << 2) | (relativeTimeEvent.timestamp >> 30));
		*buffer |= (rollover & 0xFFFF);

		// Taking the 30 bits and setting 0 to the 2 first bits.
		*(buffer + 1) = (relativeTimeEvent.timestamp & 0x3FFFFFFF);
	}
	else {
		*buffer |= (relativeTimeEvent.rollover & 0xFFFF);
		*(buffer + 1) = relativeTimeEvent.timestamp;
	}

	// Copy data.
	memcpy(buffer + 2, reinterpret_cast<const void *>(&relativeTimeEvent.data), 8);

	// Use printBinary(buffer) to debug if necessary.
}

void writeEventV2(Event const & event, int32_t * buffer) {

	Event relativeTimeEvent(event);

	*buffer = 0;
	*buffer |= ((relativeTimeEvent.crate & 0xF) << 28);
	*buffer |= ((relativeTimeEvent.board & 0x3F) << 22);
	*buffer |= ((relativeTimeEvent.channel & 0x3F) << 16);

	BoardType boardType = relativeTimeEvent.boardType();

	*buffer |= (relativeTimeEvent.rollover & 0xFFFF);
	*(buffer + 1) = relativeTimeEvent.timestamp;

	// Copy data.
	memcpy(buffer + 2, reinterpret_cast<const void *>(&relativeTimeEvent.data), 8);

	// Use printBinary(buffer) to debug if necessary.
}

void writeEvent(Event const & event, int32_t * buffer) {

	// Check the lst version to write the event.
	if (listModeContext.version == 1) {
		writeEventV1(event, buffer);
	}
	else {
		writeEventV2(event, buffer);
	}
}

}
