/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LSTDPP_EVENT_H
#define LSTDPP_EVENT_H

#include <iostream>
#include <set>
#include "CrateBoard.h"

namespace lstdpp128 {

typedef int64_t time_type;

/**
 * Structure defining the event in lstdpp.
 */
struct Event {
	uint8_t crate;
	uint8_t board;
	uint16_t channel;
	uint16_t rollover;
	uint32_t timestamp;
	uint32_t data[2];

	enum CleanType {CLEAN, NOT_CLEAN, ANTI_COMPTON, UNDEFINED};
	CleanType status;

	Event() :
		crate(0),
		board(0),
		channel(0),
		rollover(0),
		timestamp(0),
		status(UNDEFINED) {
			data[0] = 0;
			data[1] = 0;
	}

	/**
	 * Returns the board type of the event in the current list mode context.
	 */
	inline BoardType boardType() const {
		return listModeContext.crateBoard.crates[crate].boards[board].boardType;
	}

	/**
	 * Returns the time resolution of the event in the current list mode context.
	 */
	inline uint32_t timeResolution() const {
		return BoardTimeResolution[listModeContext.crateBoard.crates[crate].boards[board].boardType];
	}

	/**
	 * Gets the absolute time.
	 */
	time_type time() const;

	/**
	 * Sets the absolute time. The value is converted into rollover and timestamp.
	 */
	void setTime(time_type value);
};

/**
 * Comparator to sort events with time.
 */
struct CompareEvents {

	inline bool operator()(Event const & a, Event const & b) {
		return a.time() < b.time();
	}
};

std::string eventColumnsToString(BoardType type);
std::ostream& operator<< (std::ostream& os, Event const & e);

std::string printBinary(int32_t * buffer);

bool readEvent(Event & event, int32_t * buffer);
void writeEvent(Event const & event, int32_t * buffer);

struct EventBlock {

	static int32_t maxId;
	int32_t id;

	std::multiset<Event, CompareEvents> events;

	EventBlock() {
		id = maxId;
		maxId++;
	}

	inline time_type minTime() const {
		return events.begin()->time();
	}

	inline time_type maxTime() const {
		std::multiset<Event, CompareEvents>::const_reverse_iterator e = events.rend();
		++e;
		return e->time();
	}

	EventBlock * split();
	EventBlock * split(time_type midTime);
};

std::ostream& operator<< (std::ostream& os, EventBlock const & b);

}

#endif
