/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LSTDPP_EVENTARRAYALGORITHM_H
#define LSTDPP_EVENTARRAYALGORITHM_H

#include <iostream>
#include "BaseAlgorithm.h"
#include "EventArray.h"

namespace lstdpp128 {

template<typename Processor>
class EventArrayAlgorithm : public BaseAlgorithm {

public:
	explicit EventArrayAlgorithm();
	virtual ~EventArrayAlgorithm();

	Processor & processor();

	void process(int32_t * buffer, int32_t size, EventArray & events);
	void finish(EventArray & events);

private:
	Processor _processor;
};

template<typename Processor>
EventArrayAlgorithm<Processor>::EventArrayAlgorithm() :
	BaseAlgorithm() {
}

template<typename Processor>
EventArrayAlgorithm<Processor>::~EventArrayAlgorithm() {
}

template<typename Processor>
Processor & EventArrayAlgorithm<Processor>::processor() {
	return _processor;
}

template<typename Processor>
void EventArrayAlgorithm<Processor>::process(int32_t * buffer, int32_t size, EventArray & events) {

	// reading next block
	int32_t index = 0;
	int32_t * bufferPtr = buffer;

	while (index < size) {
		// reading event
		Event event;

		if (!readEvent(event, bufferPtr + index)) {
			_processor.processBadEvent(event);

			// jump 128 bits
			index += EVENT_SIZE_32;

			continue;
		}

		if (!_processor.checkEvent(event)) {
			_processor.processBadEvent(event);

		} else if (events.isInsertable(event.time())) {
			_processor.processEvent(event);

			EventArray::InsertResult result = events.insertEvent(event);

			if (result.process) {
				// do the process
				EventArray::array_type::iterator center = _processor.processBlock(result.begin, result.end, false);

				// update the processed center
				events.setProcessedCenter(center);
			}

		} else {
			_processor.processEventNotInserted(event);

		}

		// jump 128 bits
		index += EVENT_SIZE_32;
	}
}

template<typename Processor>
void EventArrayAlgorithm<Processor>::finish(EventArray & events) {

	events.pack();

	// in case of init state, only process the beginning of the array
	if (std::distance(events.center(), events.array().end()) < 0) {
		_processor.processBlock(events.array().begin(), events.array().end(), true);
	} else {
		// processing in two steps, because the processor cannot manage blocks
		// with size greater than half the size of the events array
		EventArray::array_type::iterator center = _processor.processBlock(events.array().begin(), events.center(), false);

		// update the processed center
		events.setProcessedCenter(center);

		_processor.processBlock(center, events.array().end(), true);
	}

	_processor.finish();
}

}

#endif
