/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LSTDPP_EVENTBLOCKARRAY_H
#define LSTDPP_EVENTBLOCKARRAY_H

#include <list>

#include "Event.h"

namespace lstdpp128 {

class EventBlockArrayIterator;

/**
 * The EventBlockArray is an array of EventBlock.
 */
class EventBlockArray {

public:
	typedef std::list<EventBlock *> array_type;

	/**
	 * The maxBlockSize is the maximum number of elements of a block and maxBlocks is the maximum number of blocks.
	 */
	EventBlockArray(int32_t maxBlockSize, int32_t maxBlocks = -1) :
		_maxBlockSize(maxBlockSize), _maxBlocks(maxBlocks), _reduced(false) {

		addEventBlock();
	}

	~EventBlockArray() {
		for (array_type::iterator b = _array.begin(); b != _array.end(); ++b) {
			delete *b;
		}
	}

	/**
	 * Returns the removed block or the null pointer.
	 */
	EventBlock * updateBlocks();
	void addEventBlock();
	bool isInsertable(time_type time) const;

	inline array_type & array() {
		return _array;
	}

	inline array_type const & array() const {
		return _array;
	}

	EventBlockArrayIterator begin() const;
	EventBlockArrayIterator end() const;

	time_type maxPreviousBlockTime(EventBlockArray::array_type::reverse_iterator const & b) const;
	EventBlockArrayIterator insertEvent(Event const & event);

private:
	int32_t _maxBlockSize;
	int32_t _maxBlocks;
	bool _reduced;
	array_type _array;
};

class EventBlockArrayIterator {

public:
	EventBlockArrayIterator(EventBlockArray::array_type const & EventBlocks, EventBlockArray::array_type::const_iterator const & b, std::multiset<Event, CompareEvents>::iterator const & e) :
		_EventBlocks(EventBlocks), _b(b), _e(e) {
	}

	inline bool first() const {
		return (_b == _EventBlocks.begin() && _e == (*_b)->events.begin());
	}

	inline bool end() const {
		return (_b == _EventBlocks.end());
	}

	inline Event & operator* () const {
		// why do we need a const_cast?
		return const_cast<Event &>(*_e);
	}

	inline bool operator== (EventBlockArrayIterator const & i) const {
		if (_b == _EventBlocks.end()) {
			return _b == i._b;
		}
		return _b == i._b && _e == i._e;
	}

	inline bool operator!= (EventBlockArrayIterator const & i) const {
		return !(*this == i);
	}

	inline EventBlockArrayIterator & operator++ () {

		if (!end()) {
			++_e;
			if (_e == (*_b)->events.end()) {
				++_b;
				if (end()) {
					return *this;
				}
				_e = (*_b)->events.begin();
			}
		}

		return *this;
	}

	inline EventBlockArrayIterator & operator-- () {

		if (!first()) {
			if (!end() && _e != (*_b)->events.begin()) {
				--_e;
				return *this;
			}

			--_b;
			std::multiset<Event, CompareEvents>::const_reverse_iterator ec = (*_b)->events.rbegin();
			// needing to advance before calling base to convert from reverse to forward iterator
			++ec;
			_e = ec.base();
		}

		return *this;
	}

private:
	EventBlockArray::array_type const & _EventBlocks;
	EventBlockArray::array_type::const_iterator _b;
	std::multiset<Event, CompareEvents>::iterator _e;
};

//void testIterator(EventBlockArray const & EventBlocks);
//void testSplitBlocks(EventBlockArray const & EventBlocks);
//void printBlocks(EventBlockArray const & EventBlocks, bool printEvents);
void checkEvents(EventBlockArray const & EventBlocks);

}

#endif
