/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Insert.h"

#include <iomanip>

using namespace std;

namespace lstdpp128 {

Insert::Insert() :
	_maxNumberOfChannels(0),
	_numberOfGoodEvents(0),
	_numberOfBadEvents(0),
	_numberOfEventsNotInserted(0) {

}

void Insert::init() {

	_maxNumberOfChannels = (1 << MAX_BITS);
	_numberOfGoodEvents = 0;
	_numberOfBadEvents = 0;
	_numberOfEventsNotInserted = 0;

	// initialise the maps
	for (int c = 0; c < listModeContext.crateBoard.nbCrates; c++) {
		for (int b = 0; b < listModeContext.crateBoard.crates[c].nbBoards; b++) {
			for (int ch = 0; ch < listModeContext.crateBoard.crates[c].boards[b].nbChannels; ch++) {
				_channelCounterInserted[c][b][ch] = 0;
				_channelCounterNotInserted[c][b][ch] = 0;
			}
		}
	}
}

Insert::~Insert() {
}

int32_t Insert::numberOfGoodEvents() const {
	return _numberOfGoodEvents;
}

int32_t Insert::numberOfBadEvents() const {
	return _numberOfBadEvents;
}

int32_t Insert::numberOfEventsNotInserted() const {
	return _numberOfEventsNotInserted;
}

const std::map<int32_t, std::map<int32_t, std::map<int32_t, int32_t> > >& Insert::channelCounterInserted() const {
	return _channelCounterInserted;
}

const std::map<int32_t, std::map<int32_t, std::map<int32_t, int32_t> > >& Insert::channelCounterNotInserted() const {
	return _channelCounterNotInserted;
}


bool Insert::checkEvent(Event const & event) {

	if (event.crate >= listModeContext.crateBoard.nbCrates) {
		return false;
	}

	if (event.board >= listModeContext.crateBoard.crates[event.crate].nbBoards) {
		return false;
	}

	if (event.channel >= listModeContext.crateBoard.crates[event.crate].boards[event.board].nbChannels) {
		return false;
	}

	return true;
}

void Insert::processBadEvent(Event const & event) {
	_numberOfBadEvents++;
}

void Insert::processEvent(Event const & event) {
	_numberOfGoodEvents++;

	_channelCounterInserted[event.crate][event.board][event.channel]++;
}

void Insert::processEvent(EventBlockArrayIterator const & e) {
	_numberOfGoodEvents++;

	_channelCounterInserted[(*e).crate][(*e).board][(*e).channel]++;
}

void Insert::processEventNotInserted(Event const & event) {
	_numberOfGoodEvents++;
	_numberOfEventsNotInserted++;

	_channelCounterNotInserted[event.crate][event.board][event.channel]++;
}

void Insert::finish() {

	cout << "number of events inserted, not inserted" << endl;

	for (map<int32_t, map<int32_t, map<int32_t, int32_t> > >::const_iterator c = _channelCounterInserted.begin(); c != _channelCounterInserted.end(); ++c) {
		cout << "crate " << c->first << endl;
		map<int32_t, map<int32_t, int32_t> > boardInserted = c->second;
		map<int32_t, map<int32_t, int32_t> > boardNotInserted = _channelCounterNotInserted[c->first];

		for (map<int32_t, map<int32_t, int32_t> >::const_iterator b = boardInserted.begin(); b != boardInserted.end(); ++b) {
			cout << "\tboard " << b->first << endl;

			map<int32_t, int32_t> channelInserted = b->second;
			map<int32_t, int32_t> channelNotInserted = boardNotInserted[b->first];

			for (map<int32_t, int32_t>::const_iterator ch = channelInserted.begin(); ch != channelInserted.end(); ++ch) {
				cout << "\t\t" << setw(5) << ch->first << " = " << setw(10) << ch->second << setw(10) << channelNotInserted[ch->first] << endl;
			}
		}
	}
}



}
