/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LiveReader.h"

#include <iostream>
#include <unistd.h>

using namespace std;

namespace lstdpp128 {

LiveReader::LiveReader(int32_t size) :
	_end(false) {
	init(size);
}

LiveReader::~LiveReader() {
}

bool LiveReader::open(std::string const & fileName) {

	int counter = 0;

	while (!_end && !_dataFile.is_open() && counter < RETRY) {
		usleep(WAIT_TIME_US);
		_dataFile.open(fileName.c_str(), ios::in | ios::binary);

		if (_dataFile.is_open()) {

			_dataFile.seekg(0, ios::end);
			_fileSize = _dataFile.tellg();
			_dataFile.seekg(0, ios::beg);

			// Read the 4 fixed elements of the header.
			uint32_t nbEvents = 0;
			uint32_t nbBoards = 0;

			_dataFile.read(reinterpret_cast<char *>(&_version), sizeof(int32_t));
			_dataFile.read(reinterpret_cast<char *>(&_timeBase), sizeof(int32_t));
			_dataFile.read(reinterpret_cast<char *>(&nbEvents), sizeof(int32_t));
			_dataFile.read(reinterpret_cast<char *>(&nbBoards), sizeof(int32_t));

			// Read the board structure.
			_boardHeader = new int32_t[nbBoards];
			_dataFile.read(reinterpret_cast<char *>(_boardHeader), nbBoards * sizeof(uint32_t));

			// Read the list mode context.
			_totalNumberOfBoards = readListModeContextFromBinary(listModeContext, _version, _timeBase, nbEvents, nbBoards, (uint32_t*)_boardHeader);
		}

		++counter;
	}

	if (_end) {
		return false;
	}

	if (counter == RETRY) {
		cerr << "problem reading " << fileName << endl;
		return false;
	}

	return true;
}

int32_t LiveReader::read() {

	FileSize blockSize = getBlockSizeInBytes();

	//cout << "block size " << blockSize << endl;

	// reading next block
	while (!_end && blockSize < _sizeInBytes) {
		// waiting before getting the block
		usleep(WAIT_TIME_US);
		blockSize = getBlockSizeInBytes();
	}

	if (_end) {
		int retry = 0;

		cout << "retrying LiveReader" << endl;

		while (retry < 500 && blockSize == (FileSize)0) {
			// waiting before getting the block
			usleep(WAIT_TIME_US);
			blockSize = getBlockSizeInBytes();

			++retry;
		}

		cout << "found block " << blockSize << endl;
	}

	//cout << "reading file " << blockSize << " elements = " << (blockSize / 4) << " size = " << _size << endl;

	if (blockSize < _sizeInBytes) {
		_dataFile.read((char *)_buffer, blockSize);
		return (blockSize / 4);
	}

	_dataFile.read((char *)_buffer, _sizeInBytes);

	return _size;
}

BaseReader::FileSize LiveReader::getBlockSizeInBytes() {

	FileSize currentPosition = _dataFile.tellg();

	_dataFile.seekg(0, ios::end);
	FileSize size = _dataFile.tellg() - currentPosition;
	_dataFile.seekg(currentPosition, ios::beg);

	return size;
}

void LiveReader::stop() {
	_end = true;
}

}
