/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "../../algorithms/nomad/Coincidence.h"
#include <iomanip>
#include <cmath>

using namespace std;

namespace lstdpp128 {

int32_t binomialCoefficient(int32_t n, int32_t k) {
	if (n >= 0 && k == 0) {
		return 1;
	}

	if (n == 0 && k > 0) {
		return 0;
	}

	return binomialCoefficient(n - 1, k - 1) + binomialCoefficient(n - 1, k);
}

Coincidence::Coincidence() :
	_totalNbChannels(0),
	_coincidenceRates(0),
	_counters(0),
	_coincidenceMatrixResolution(0),
	_coincidenceMatrixXSize(0),
	_coincidenceMatrix(0),
	_maxOrder(0),
	_channelCounters(0),
	_channelRates(0),
	_maxNumberOfEnergyChannels(0),
	_coincidenceDT(0),
	_cleanDT(0),
	_outMemblock(0),
	_numberOfGoodEvents(0),
	_numberOfBadEvents(0),
	_numberOfEventsNotInserted(0),
	_maxTime(0),
	_numberOfTruncatedCoincidences(0) {
}

void Coincidence::init(int32_t* headerMemoryBlock, uint32_t version, uint32_t timeBase, uint32_t numberOfBoards, int16_t coincidenceMatrixResolution, const std::vector<int16_t>& detectorADCs, time_type coincidenceDT, time_type cleanDT, int32_t maxBlockSize, std::string const & outFileName) {

	_coincidenceDT = coincidenceDT;
	_cleanDT = cleanDT;

	// Generate the channel offset and the total size of the histograms
	BoardIterator b;
	int32_t currentNbChannels = (*b).nbChannels;

	++b;
	while (!b.end()) {
		_channelOffset[b.crate()][b.board()] = currentNbChannels;
		cout << "channel offset " << (int)b.crate() << ", " << (int)b.board() << " = " << _channelOffset[b.crate()][b.board()] << endl;
		currentNbChannels += (*b).nbChannels;

		++b;
	}

	// The total number of channels is the current number of channels
	_totalNbChannels = currentNbChannels;

	_coincidenceRates = new double[4 * 7];
	for (int i = 0; i < 4 * 7; ++i) {
		_coincidenceRates[i] = 0.0;
	}

	_counters = new uint64_t[4 * 7];
	memset((void *)_counters, 0, 4 * 7 * sizeof(uint64_t));

	_coincidenceMatrixResolution = coincidenceMatrixResolution;

	if (_coincidenceMatrixResolution > 0) {
		_coincidenceMatrixXSize = (1 << coincidenceMatrixResolution);
		_coincidenceMatrix = new int32_t[_coincidenceMatrixXSize * _coincidenceMatrixXSize];
		memset((void *)_coincidenceMatrix, 0, _coincidenceMatrixXSize * _coincidenceMatrixXSize * sizeof(int32_t));
	}

	// organize detector IDs
	int16_t detectorIDMax = 0;
	for (vector<int16_t>::const_iterator d = detectorADCs.begin(); d != detectorADCs.end(); d++) {
		if (*d < 0) {
			int16_t detectorID = -1 - (*d);
			_detectorChannels.push_back(detectorID);
			_antiComptonChannels.push_back(true);

			if (detectorID > detectorIDMax) {
				detectorIDMax = detectorID;
			}

		} else {
			int16_t detectorID = *d;
			_detectorChannels.push_back(detectorID);
			_antiComptonChannels.push_back(false);

			if (detectorID > detectorIDMax) {
				detectorIDMax = detectorID;
			}
		}
	}

	_lastAntiComptonTimeChannels.resize(detectorIDMax + 1);

	// Allocate the channel counters and rates
	_channelCounters = new uint64_t[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelCounters[i] = 0;
	}

	_channelRates = new double[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelRates[i] = 0.0;
	}

	_maxNumberOfEnergyChannels = (1 << MAX_BITS);

	_outFileName = outFileName;

	if (_outFileName != "") {
		string tempOutFileName = outFileName + ".temp";
		_outDataFile.open(tempOutFileName.c_str());

		uint32_t nbEvents = 0;

		// Write the header.
		_outDataFile.write((char *)&version, sizeof(uint32_t));
		_outDataFile.write((char *)&timeBase, sizeof(uint32_t));
		_outDataFile.write((char *)&nbEvents, sizeof(uint32_t));
		_outDataFile.write((char *)&numberOfBoards, sizeof(uint32_t));

		_outDataFile.write((char *)headerMemoryBlock, numberOfBoards * sizeof(uint32_t));
	}

	_numberOfGoodEvents = 0;
	_numberOfBadEvents = 0;
	_numberOfEventsNotInserted = 0;
	_numberOfTruncatedCoincidences = 0;

	// event encoded onto 4 int32_t, the output memory block can be twice maxBlockSize following the EventArray size
	_outMemblock = new int32_t[maxBlockSize * 4 * 2];

	// initialising the coincidence array
	updateCoincidenceArray(12);

	// reserving 8 elements
	_eventCoincidenceArray.reserve(8);
}

Coincidence::~Coincidence() {
	delete [] _coincidenceRates;
	delete [] _counters;
	delete [] _channelCounters;
	delete [] _channelRates;
	delete [] _outMemblock;
}

int32_t Coincidence::numberOfGoodEvents() const {
	return _numberOfGoodEvents;
}

int32_t Coincidence::numberOfBadEvents() const {
	return _numberOfBadEvents;
}

int32_t Coincidence::numberOfEventsNotInserted() const {
	return _numberOfEventsNotInserted;
}

int32_t Coincidence::numberOfTruncatedCoincidences() const {
	return _numberOfTruncatedCoincidences;
}

int16_t Coincidence::maxOrder() const {
	return _maxOrder;
}

std::vector<Event> Coincidence::maxEventCoincidenceArray() const {
	return _maxEventCoincidenceArray;
}

bool Coincidence::checkEvent(Event const & event) {
	return true;
}

void Coincidence::processBadEvent(Event const & event) {
	_numberOfBadEvents++;
}

void Coincidence::processEvent(Event const & event) {
	_numberOfGoodEvents++;

	// updating ADC Counter
	int32_t index = _channelOffset[event.crate][event.board] + event.channel;

	_channelCounters[index]++;
	_channelRates[index] = pow(10, listModeContext.timeBase) * (double)_channelCounters[index] / (double)event.time();
}

void Coincidence::processEvent(EventBlockArrayIterator const & e) {
	_numberOfGoodEvents++;

	int32_t index = _channelOffset[(*e).crate][(*e).board] + (*e).channel;

	// Update the counter and rates
	_channelCounters[index]++;
	_channelRates[index] = pow(10, listModeContext.timeBase) * (double)_channelCounters[index] / (double)(*e).time();
}

void Coincidence::processEventNotInserted(Event const & event) {
	_numberOfGoodEvents++;
	_numberOfEventsNotInserted++;
}

int16_t Coincidence::detector(Event const & event) {
	int32_t index = _channelOffset[event.crate][event.board] + event.channel;
	return _detectorChannels[index];
}

bool Coincidence::antiCompton(Event const & event) {
	int32_t index = _channelOffset[event.crate][event.board] + event.channel;
	return _antiComptonChannels[index];
}

void Coincidence::setLastAntiComptonTime(Event const & event) {
	_lastAntiComptonTimeChannels[detector(event)] = event.time();
}

time_type Coincidence::lastAntiComptonTime(Event const & event) {
	return _lastAntiComptonTimeChannels[detector(event)];
}

void Coincidence::updateCoincidenceArray(int16_t order) {

	for (int32_t n = _coincidenceArray.size(); n < order + 1; ++n) {
		_coincidenceArray.push_back(vector<int32_t>());
		vector<int32_t> & array = _coincidenceArray.back();
		array.resize(n + 1);

		for (int32_t k = 0; k < n + 1; ++k) {
			array[k] = binomialCoefficient(n, k);
		}
	}
}

int32_t Coincidence::numberOfCoincidences(int16_t order, int16_t k) {

	if (_coincidenceArray.size() <= order) {
		return 0;
	}
	return _coincidenceArray[order][k];
}

int16_t Coincidence::index(int16_t mode, int16_t order) {
	return mode * 7 + order - 2;
}

int32_t Coincidence::bitshift(Event const & event) const {
	return MAX_BITS - eventEnergyResolution(event);
}

void Coincidence::addCoincidence(int16_t order, bool isClean, bool isDetector) {

	// Events are supposed to have energy.

	//cout << "addCoincidence " << order << " is clean " << isClean << " is detector " << isDetector << endl;

	// updating max order
	if (order > _maxOrder) {
		_maxOrder = order;
		_maxEventCoincidenceArray = _eventCoincidenceArray;
	}

	int16_t processedOrder = min(order, (int16_t)8);

	// updating rates
	for (int16_t k = processedOrder; k > 1; --k) {

		int32_t n = numberOfCoincidences(order, k);

		//cout << "n(" << order << ", " << k << ") = " << n << endl;

		_counters[index(GLOBAL, k)] += n;

		if (isClean) {
			_counters[index(GLOBAL_CLEAN, k)] += n;
		}

		if (isDetector) {
			_counters[index(DETECTOR, k)] += n;
		}

		if (isDetector && isClean) {
			_counters[index(DETECTOR_CLEAN, k)] += n;
		}
	}

	// update the coincidence matrix only if resolution is not zero
	if (_coincidenceMatrixResolution > 0) {

		// updating coincidence matrix
		// we update the (a, b) and (b, a) couple of events
		for (int16_t i = 0; i < processedOrder; ++i) {
			for (int16_t j = 0; j < processedOrder; ++j) {
				if (j != i) {
					Event const & eventA = _eventCoincidenceArray[i];
					int32_t channelA = (eventEnergy(eventA) >> bitshift(eventA));

					Event const & eventB = _eventCoincidenceArray[j];
					int32_t channelB = (eventEnergy(eventB) >> bitshift(eventB));

					++_coincidenceMatrix[channelA * _coincidenceMatrixXSize + channelB];
				}
			}
		}
	}

	_eventCoincidenceArray.clear();
}

void Coincidence::processEventInCoincidence(Event const & event, int16_t order, std::set<int16_t> & detectors, bool & isClean, bool & isDetector) {

	// updating current coincidence
	_eventCoincidenceArray.push_back(event);

	if (!antiCompton(event)) {

		isClean &= (event.status == Event::CLEAN);
		int16_t detectorID = detector(event);

		if (isDetector && detectors.find(detectorID) != detectors.end()) {
			isDetector = false;
		} else {
			detectors.insert(detectorID);
		}
	}
}

void Coincidence::updateRates() {

	//cout << "time = " << _maxTime << " (" << (double)_maxTime * 1.0E-8 << ")" << endl;

	for (int i = 0; i < 4 * 7; ++i) {
		_coincidenceRates[i] = pow(10, listModeContext.timeBase) * (double)_counters[i] / (double)_maxTime;
	}

	//cout << "rate global = " << _rates[0] << endl;
}

void Coincidence::finish() {

	if (_outFileName != "") {

		// Write the number of events.
		// The pointer is 2 uint32 after the beginning of the file.
		_outDataFile.seekp(2 * sizeof(uint32_t), ios::beg);
		_outDataFile.write((char *)&_numberOfGoodEvents, sizeof(uint32_t));

		_outDataFile.close();

		string tempOutFileName = _outFileName + ".temp";
		rename(tempOutFileName.c_str(), _outFileName.c_str());
	}
}

void printRate(double * rate) {

	for (int i = 0; i < 7; ++i) {
		cout << rate[i] << " ";
	}
	cout << endl;
}

void Coincidence::printRates() {

	cout << "crystal        rate = ";
	printRate(_coincidenceRates + 0);
	cout << "crystal clean  rate = ";
	printRate(_coincidenceRates + 7);
	cout << "detector       rate = ";
	printRate(_coincidenceRates + 7 * 2);
	cout << "detector clean rate = ";
	printRate(_coincidenceRates + 7 * 3);

	cout << "channel rates" << endl;

	BoardIterator b;
	int32_t channel = 0;
	while (!b.end()) {

		cout << "crate " << (int)b.crate() << " board " << (int)b.board() << endl;
		for (int32_t c = 0; c < (*b).nbChannels; ++c) {
			cout << "\t" << setw(5) << c << setw(10) << _channelRates[channel] << endl;
			++channel;
		}
		cout << endl;

		++b;
	}
}

int32_t Coincidence::getTotalNumberOfChannels() const {
	return _totalNbChannels;
}

double * Coincidence::getChannelRates() const {
	return _channelRates;
}

int32_t Coincidence::getNumberOfCoincidenceRates() const {
	return 4 * 7;
}

double * Coincidence::getCoincidenceRates() const {
	return _coincidenceRates;
}

}
