/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LSTDPP_COINCIDENCE_H
#define LSTDPP_COINCIDENCE_H

#include <cstring>
#include <algorithm>
#include <fstream>
#include <vector>
#include <map>
#include <time.h>

#include "../../Event.h"
#include "../../EventData.h"
#include "../../EventBlockArray.h"

namespace lstdpp128 {

class Coincidence {

public:
	Coincidence();
	~Coincidence();

	void init(int32_t* headerMemoryBlock, uint32_t version, uint32_t timeBase, uint32_t numberOfBoards, int16_t coincidenceMatrixResolution, const std::vector<int16_t>& detectorADCs, time_type coincidenceDT, time_type cleanDT, int32_t maxBlockSize, std::string const & outFileName);

	int32_t numberOfGoodEvents() const;
	int32_t numberOfBadEvents() const;
	int32_t numberOfEventsNotInserted() const;
	int32_t numberOfTruncatedCoincidences() const;
	int16_t maxOrder() const;
	std::vector<Event> maxEventCoincidenceArray() const;
	int16_t detector(Event const & event);
	bool antiCompton(Event const & event);

	/**
	 * Functions to implement.
	 */
	bool checkEvent(Event const & event);
	void processBadEvent(Event const & event);
	void processEvent(Event const & event);
	void processEvent(EventBlockArrayIterator const & e);
	void processEventNotInserted(Event const & event);

	template<typename Iterator>
	Iterator processBlock(Iterator first, Iterator last, bool lastBlock);
	void finish();
	void printRates();

	int32_t getTotalNumberOfChannels() const;
	double * getChannelRates() const;
	int32_t getNumberOfCoincidenceRates() const;
	double * getCoincidenceRates() const;

private:
	void updateCoincidenceArray(int16_t order);
	int32_t numberOfCoincidences(int16_t order, int16_t k);
	static int16_t index(int16_t mode, int16_t order);
	int32_t bitshift(Event const & event) const;
	void setLastAntiComptonTime(Event const & event);
	time_type lastAntiComptonTime(Event const & event);

	void addCoincidence(int16_t order, bool isClean, bool isDetector);
	void processEventInCoincidence(Event const & event, int16_t order, std::set<int16_t> & detectors, bool & isClean, bool & isDetector);

	template<typename Iterator>
	Iterator calculateCleanEvents(Iterator first, Iterator last, bool lastBlock);

	template<typename Iterator>
	Iterator calculateCoincidences(Iterator first, Iterator last, bool lastBlock);

	void updateRates();

	template<typename Iterator>
	void writeBlock(Iterator first, Iterator last);

	// crate, board - channel offset
	std::map<int32_t, std::map<int32_t, int32_t> > _channelOffset;

	int32_t _totalNbChannels;
	double * _coincidenceRates;
	uint64_t * _counters;
	int16_t _coincidenceMatrixResolution;
	int32_t _coincidenceMatrixXSize;
	int32_t * _coincidenceMatrix;
	std::vector<Event> _eventCoincidenceArray;
	std::vector<Event> _maxEventCoincidenceArray;
	std::vector<int16_t> _detectorChannels;
	std::vector<bool> _antiComptonChannels;
	std::vector<time_type> _lastAntiComptonTimeChannels;
	std::vector<std::vector<int32_t> > _coincidenceArray;
	int16_t _maxOrder;
	uint64_t * _channelCounters;
	double * _channelRates;
	int32_t _maxNumberOfEnergyChannels;
	time_type _coincidenceDT;
	time_type _cleanDT;

	int32_t * _outMemblock;
	std::string _outFileName;
	std::ofstream _outDataFile;

	int32_t _numberOfGoodEvents;
	int32_t _numberOfBadEvents;
	int32_t _numberOfEventsNotInserted;
	time_type _maxTime;

	int32_t _numberOfTruncatedCoincidences;

	static const int16_t MAX_BITS = 15;
	static const int16_t GLOBAL = 0;
	static const int16_t GLOBAL_CLEAN = 1;
	static const int16_t DETECTOR = 2;
	static const int16_t DETECTOR_CLEAN = 3;
};

template<typename Iterator>
Iterator Coincidence::processBlock(Iterator first, Iterator last, bool lastBlock) {

	last = calculateCleanEvents(first, last, lastBlock);
	last = calculateCoincidences(first, last, lastBlock);
	updateRates();
	writeBlock(first, last);

	return last;
}

template<typename Iterator>
Iterator Coincidence::calculateCleanEvents(Iterator first, Iterator last, bool lastBlock) {

	Iterator e = first;
	Iterator processedLast = first;

	// iterating events
	while (e != last) {

		// forcing const_cast because iterators on multiset return const references
		Event & event = const_cast<Event &>(*e);
		int16_t detectorID = detector(event);

		if (antiCompton(event)) {
			setLastAntiComptonTime(event);
			event.status = Event::ANTI_COMPTON;

		} else {
			// change dT into special dT for anti-compton test
			if (event.time() - lastAntiComptonTime(event) <= _cleanDT) {
				event.status = Event::NOT_CLEAN;

			} else {
				Iterator b = e;
				++b;
				bool foundStatus = false;
				while (b != last) {
					if ((*b).time() - event.time() > _cleanDT) {
						event.status = Event::CLEAN;
						foundStatus = true;
						break;
					} else if (antiCompton(*b) && detector(*b) == detectorID) {
						event.status = Event::NOT_CLEAN;
						foundStatus = true;
						break;
					}

					++b;
				}

				if (!foundStatus && b == last) {
					processedLast = e;
					break;
				}
			}
		}

		++e;
	}

	if (!lastBlock && processedLast != first) {
		//std::cout << "distance with last = " << std::distance(processedLast, last) << std::endl;
		return processedLast;
	}

	// in case of last block, considering that last events are clean
	while (e != last) {
		// forcing const_cast because iterators on multiset return const references
		Event & event = const_cast<Event &>(*e);

		if (antiCompton(event)) {
			event.status = Event::ANTI_COMPTON;
		} else {
			event.status = Event::CLEAN;
		}

		++e;
	}

	return last;
}

template<typename Iterator>
Iterator Coincidence::calculateCoincidences(Iterator first, Iterator last, bool lastBlock) {

	int16_t order = 0;
	bool isClean = true;
	bool isDetector = true;
	std::set<int16_t> detectors;
	Iterator processedLast = first;
	Iterator e = first;
	time_type endTime = (*e).time() + _coincidenceDT;

	processEventInCoincidence(*e, order, detectors, isClean, isDetector);

	++order;
	++e;

	// iterating events
	while (e != last) {

		Event const & event = *e;

		_maxTime = event.time();

		if (event.time() > endTime) {

			// add current coincidence
			addCoincidence(order, isClean, isDetector);

			// new coincidence interval
			order = 0;
			isClean = true;
			isDetector = true;
			detectors.clear();
			processedLast = e;
			endTime = event.time() + _coincidenceDT;
		}

		processEventInCoincidence(event, order, detectors, isClean, isDetector);

		++order;
		++e;
	}

	if (!lastBlock && processedLast != first) {
		return processedLast;
	}

	if (!lastBlock) {
		_numberOfTruncatedCoincidences++;
	}

	// add the last coincidence
	addCoincidence(order, isClean, isDetector);

	// returning last in case of first coincidence
	return last;
}

template<typename Iterator>
void Coincidence::writeBlock(Iterator first, Iterator last) {

	if (_outFileName != "") {
		int32_t * buffer = _outMemblock;
		int32_t index = 0;

		// iterating events
		for (Iterator e = first; e != last; ++e) {
			writeEvent(*e, buffer + index);
			index += 2;
		}

		_outDataFile.write((char *)_outMemblock, index * sizeof(int32_t));
	}
}

}

#endif
