/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "../../algorithms/nomad/Histogram.h"

#include <iostream>
#include <iomanip>
#include <cmath>

using namespace std;

namespace lstdpp128 {

Histogram::Histogram() :
	_totalNbChannels(0),
	_channelCounters(0),
	_channelRates(0),
	_totalHistogramSize(0),
	_histograms(0),
	_triggerCrate(0),
	_triggerBoard(0),
	_triggerChannel(0),
	_delay(0),
	_dT(0),
	_maxBlockSize(0),
	_outMemblock(0),
	_numberOfGoodEvents(0),
	_numberOfBadEvents(0),
	_numberOfEventsNotInserted(0),
	_numberOfEventsInConflict(0) {
}

void Histogram::init(int32_t* headerMemoryBlock, uint32_t version, uint32_t timeBase, uint32_t numberOfBoards, uint8_t triggerCrate, uint8_t triggerBoard, uint16_t triggerChannel, time_type delay, time_type dT, int32_t maxBlockSize, std::string const & outFileName) {

	_triggerCrate = triggerCrate;
	_triggerBoard = triggerBoard;
	_triggerChannel = triggerChannel;

	_delay = delay;
	_dT = dT;

	// Generate the channel offset and the total size of the histograms
	BoardIterator b;
	int32_t currentNbChannels = (*b).nbChannels;
	int32_t currentHistogramSize = (*b).nbChannels * (1 << boardEnergyResolution((*b).boardType));

	++b;
	while (!b.end()) {
		_channelOffset[b.crate()][b.board()] = currentNbChannels;
		cout << "channel offset " << (int)b.crate() << ", " << (int)b.board() << " = " << _channelOffset[b.crate()][b.board()] << endl;
		currentNbChannels += (*b).nbChannels;

		_histogramOffset[b.crate()][b.board()] = currentHistogramSize;
		cout << "histogram offset " << (int)b.crate() << ", " << (int)b.board() << " = " << _histogramOffset[b.crate()][b.board()] << endl;

		currentHistogramSize += (*b).nbChannels * (1 << boardEnergyResolution((*b).boardType));

		++b;
	}

	// The total number of channels is the current number of channels
	_totalNbChannels = currentNbChannels;

	// The total histogram size is the current histogram size
	_totalHistogramSize = currentHistogramSize;

	// Allocate the channel counters and rates
	_channelCounters = new uint64_t[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelCounters[i] = 0;
	}

	_channelRates = new double[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelRates[i] = 0.0;
	}

	// The total histogram size is the current size of the histograms
	_histograms = new int32_t[_totalHistogramSize];
	for (int32_t i = 0; i < _totalHistogramSize; ++i) {
		_histograms[i] = 0;
	}

	_numberOfGoodEvents = 0;
	_numberOfBadEvents = 0;
	_numberOfEventsNotInserted = 0;
	_numberOfEventsInConflict = 0;

	// Manage output file
	_outFileName = outFileName;

	if (_outFileName != "") {
		string tempOutFileName = outFileName + ".temp";
		_outDataFile.open(tempOutFileName.c_str());

		uint32_t nbEvents = 0;

		// Write the header.
		_outDataFile.write((char *)&version, sizeof(uint32_t));
		_outDataFile.write((char *)&timeBase, sizeof(uint32_t));
		_outDataFile.write((char *)&nbEvents, sizeof(uint32_t));
		_outDataFile.write((char *)&numberOfBoards, sizeof(uint32_t));

		_outDataFile.write((char *)headerMemoryBlock, numberOfBoards * sizeof(uint32_t));
	}

	_maxBlockSize = maxBlockSize;
	_outMemblock = new int32_t[maxBlockSize * 4];
}

Histogram::~Histogram() {
	delete [] _channelCounters;
	delete [] _channelRates;
	delete [] _histograms;
	delete [] _outMemblock;
}

int32_t Histogram::numberOfGoodEvents() const {
	return _numberOfGoodEvents;
}

int32_t Histogram::numberOfBadEvents() const {
	return _numberOfBadEvents;
}

int32_t Histogram::numberOfEventsNotInserted() const {
	return _numberOfEventsNotInserted;
}

int32_t Histogram::numberOfEventsInConflict() const {
	return _numberOfEventsInConflict;
}

bool Histogram::checkEvent(Event const & event) {
	return true;
}

bool Histogram::isTrigger(Event const & event) const {
	return (event.crate == _triggerCrate
			&& event.board == _triggerBoard
			&& event.channel == _triggerChannel);
}

int32_t Histogram::bitshift(Event const & event) const {
	return MAX_BITS - eventEnergyResolution(event);
}

void Histogram::incrementHistogram(Event const & event) {

	// Only process events that have energy.
	if (eventHasEnergy(event)) {

		int32_t offset = _histogramOffset[event.crate][event.board];
		int32_t energySize = (1 << eventEnergyResolution(event));
		int32_t energyChannel = (eventEnergy(event) >> bitshift(event));

		_histograms[offset + event.channel * energySize + energyChannel]++;
	}
}

void Histogram::processEvent(EventBlockArrayIterator const & e) {

	_numberOfGoodEvents++;

	int32_t index = _channelOffset[(*e).crate][(*e).board] + (*e).channel;

	// Update the counter and rates
	_channelCounters[index]++;
	_channelRates[index] = pow(10, listModeContext.timeBase) * (double)_channelCounters[index] / (double)(*e).time();

	// Check if e is a trigger event
	if (isTrigger(*e)) {
		processTrigger(e);
	} else {
		processNonTrigger(e);
	}
}

void Histogram::processTrigger(EventBlockArrayIterator const & e) {

	// storing the time
	time_type triggerTime = (*e).time();

	// checking if the trigger is valid
	EventBlockArrayIterator ec = e;

	bool found = false;
	if (!ec.first()) {
		--ec;
		// searching for a trigger
		while ((*ec).time() + _delay + _dT  >= triggerTime) {

			if (isTrigger(*ec)) {
				found = true;
				break;
			}

			if (ec.first()) {
				break;
			}

			--ec;
		}
	}

	if (found) {
		_numberOfEventsInConflict++;
		return;
	}

	// updating the histogram
	incrementHistogram(*e);

	//cout << "begin with " << (*e) << endl;

	// moving to the beginning of the interval
	EventBlockArrayIterator et = e;

	if (_delay < 0) {

		while ((*et).time() > triggerTime + _delay) {

			if (et.first()) {
				break;
			}
			--et;
		}

	} else {

		while (!et.end() && (*et).time() < triggerTime + _delay) {
			++et;
		}
	}

	// iterating
	while (!et.end() && (*et).time() <= triggerTime + _delay + _dT) {

		//cout << "-> " << (*e) << endl;

		// counting when time is in interval [trigger + delay, trigger + delay + dT]
		// the test is maintained because the iteration for finding the beginning of the interval
		// does not ensure the following test
		if ((*et).time() >= triggerTime + _delay) {

			if (!isTrigger(*et)) {
				incrementHistogram(*et);

			} else {
				// why comment???
				//_numberOfEventsInConflict++;
				//break;
			}
		}

		++et;
	}
}

void Histogram::processNonTrigger(EventBlockArrayIterator const & e) {

	// storing the event
	time_type eventTime = (*e).time();

	// searching for a previous trigger
	EventBlockArrayIterator ep = e;

	while (eventTime <= (*ep).time() + _delay + _dT) {

		if (isTrigger(*ep)) {

			// counting when time is in interval [trigger + delay, trigger + delay + dT]
			if (eventTime >= (*ep).time() + _delay) {
				incrementHistogram(*ep);
			}
			break;
		}

		if (ep.first()) {
			break;
		}

		--ep;
	}

	// searching for a next trigger
	EventBlockArrayIterator en = e;

	while (!en.end() && eventTime >= (*en).time() + _delay) {

		if (isTrigger(*en)) {

			// counting when time is in interval [trigger + delay, trigger + delay + dT]
			if (eventTime <= (*en).time() + _delay + _dT) {
				incrementHistogram(*en);
			}
			break;
		}

		++en;
	}

}

void Histogram::processBadEvent(Event const & event) {
	_numberOfBadEvents++;
}

void Histogram::processEventNotInserted(Event const & event) {
	_numberOfGoodEvents++;
	_numberOfEventsNotInserted++;
}

void Histogram::finish() {

	if (_outFileName != "") {

		// Write the number of events.
		// The pointer is 2 uint32 after the beginning of the file.
		_outDataFile.seekp(2 * sizeof(uint32_t), ios::beg);
		_outDataFile.write((char *)&_numberOfGoodEvents, sizeof(uint32_t));

		_outDataFile.close();

		string tempOutFileName = _outFileName + ".temp";
		rename(tempOutFileName.c_str(), _outFileName.c_str());
	}
}

void Histogram::printRates() {

	cout << "channel rates" << endl;

	BoardIterator b;
	int32_t channel = 0;
	while (!b.end()) {

		cout << "crate " << (int)b.crate() << " board " << (int)b.board() << endl;
		for (int32_t c = 0; c < (*b).nbChannels; ++c) {
			cout << "\t" << setw(5) << c << setw(10) << _channelRates[channel] << endl;
			++channel;
		}
		cout << endl;

		++b;
	}
}

int32_t Histogram::getTotalNumberOfChannels() const {
	return _totalNbChannels;
}

double * Histogram::getChannelRates() const {
	return _channelRates;
}

int32_t Histogram::getTotalHistogramSize() const {
	return _totalHistogramSize;
}

int32_t * Histogram::getHistograms() const {
	return _histograms;
}

}
