/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "../../algorithms/nomad/Rates.h"

#include <iostream>
#include <iomanip>
#include <cmath>

using namespace std;

namespace lstdpp128 {

Rates::Rates() :
	_totalNbChannels(0),
	_channelCounters(0),
	_channelRates(0),
	_maxBlockSize(0),
	_outMemblock(0),
	_numberOfGoodEvents(0),
	_numberOfBadEvents(0),
	_numberOfEventsNotInserted(0),
	_numberOfEventsInConflict(0) {
}

void Rates::init(int32_t* headerMemoryBlock, uint32_t version, uint32_t timeBase, uint32_t numberOfBoards, int32_t maxBlockSize, std::string const & outFileName) {

	// Generate the channel offset and the total size of the histograms
	BoardIterator b;
	int32_t currentNbChannels = (*b).nbChannels;
	//int32_t currentRatesSize = (*b).nbChannels * (1 << BoardDataResolution[(*b).boardType]);

	++b;
	while (!b.end()) {
		_channelOffset[b.crate()][b.board()] = currentNbChannels;
		cout << "channel offset " << (int)b.crate() << ", " << (int)b.board() << " = " << _channelOffset[b.crate()][b.board()] << endl;
		currentNbChannels += (*b).nbChannels;

		//currentRatesSize += (*b).nbChannels * (1 << BoardDataResolution[(*b).boardType]);

		++b;
	}

	// The total number of channels is the current number of channels
	_totalNbChannels = currentNbChannels;

	// Allocate the channel counters and rates
	_channelCounters = new uint64_t[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelCounters[i] = 0;
	}

	_channelRates = new double[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelRates[i] = 0.0;
	}

	_numberOfGoodEvents = 0;
	_numberOfBadEvents = 0;
	_numberOfEventsNotInserted = 0;
	_numberOfEventsInConflict = 0;

	// Manage output file
	_outFileName = outFileName;

	if (_outFileName != "") {
		string tempOutFileName = outFileName + ".temp";
		_outDataFile.open(tempOutFileName.c_str());

		uint32_t nbEvents = 0;

		// Write the header.
		_outDataFile.write((char *)&version, sizeof(uint32_t));
		_outDataFile.write((char *)&timeBase, sizeof(uint32_t));
		_outDataFile.write((char *)&nbEvents, sizeof(uint32_t));
		_outDataFile.write((char *)&numberOfBoards, sizeof(uint32_t));

		_outDataFile.write((char *)headerMemoryBlock, numberOfBoards * sizeof(uint32_t));
	}

	_maxBlockSize = maxBlockSize;
	_outMemblock = new int32_t[maxBlockSize * 4];
}

Rates::~Rates() {
	delete [] _channelCounters;
	delete [] _channelRates;
	delete [] _outMemblock;
}

int32_t Rates::numberOfGoodEvents() const {
	return _numberOfGoodEvents;
}

int32_t Rates::numberOfBadEvents() const {
	return _numberOfBadEvents;
}

int32_t Rates::numberOfEventsNotInserted() const {
	return _numberOfEventsNotInserted;
}

int32_t Rates::numberOfEventsInConflict() const {
	return _numberOfEventsInConflict;
}

bool Rates::checkEvent(Event const & event) {
	return true;
}

void Rates::processEvent(EventBlockArrayIterator const & e) {

	_numberOfGoodEvents++;

	int32_t index = _channelOffset[(*e).crate][(*e).board] + (*e).channel;

	// Update the counter and rates
	_channelCounters[index]++;

	// Convert to Hz using time base.
	_channelRates[index] = pow(10, listModeContext.timeBase) * (double)_channelCounters[index] / (double)(*e).time();
}

void Rates::processBadEvent(Event const & event) {
	_numberOfBadEvents++;
}

void Rates::processEventNotInserted(Event const & event) {
	_numberOfGoodEvents++;
	_numberOfEventsNotInserted++;
}

void Rates::finish() {

	if (_outFileName != "") {

		// Write the number of events.
		// The pointer is 2 uint32 after the beginning of the file.
		_outDataFile.seekp(2 * sizeof(uint32_t), ios::beg);
		_outDataFile.write((char *)&_numberOfGoodEvents, sizeof(uint32_t));

		_outDataFile.close();

		string tempOutFileName = _outFileName + ".temp";
		rename(tempOutFileName.c_str(), _outFileName.c_str());
	}
}

void Rates::printRates() {

	cout << "channel rates" << endl;

	BoardIterator b;
	int32_t channel = 0;
	while (!b.end()) {

		cout << "crate " << (int)b.crate() << " board " << (int)b.board() << endl;
		for (int32_t c = 0; c < (*b).nbChannels; ++c) {
			cout << "\t" << setw(5) << c << setw(10) << _channelRates[channel] << endl;
			++channel;
		}
		cout << endl;

		++b;
	}
}

int32_t Rates::getTotalNumberOfChannels() const {
	return _totalNbChannels;
}

double * Rates::getChannelRates() const {
	return _channelRates;
}

}
