/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CoincidenceOffline.h"
#include <iomanip>
#include <cmath>

using namespace std;

namespace lstdpp128 {

const double CoincidenceOffline::_energyMax=10000;

int32_t binomialCoefficient(int32_t n, int32_t k) {
	if (n >= 0 && k == 0) {
		return 1;
	}

	if (n == 0 && k > 0) {
		return 0;
	}

	return binomialCoefficient(n - 1, k - 1) + binomialCoefficient(n - 1, k);
}

CoincidenceOffline::CoincidenceOffline() :
		_maxNumberOfEnergyChannels(0),
		_channelCounters(0),
		_channelRates(0),
		_numberOfGoodEvents(0),
		_numberOfBadEvents(0),
		_numberOfEventsNotInserted(0),
		_maxTime(0),
		_numberOfTruncatedCoincidences(0) {
}

void CoincidenceOffline::init(std::vector<int16_t> const & detectorADCs, time_type coincidenceDT, time_type cleanDT, int mult, int32_t maxBlockSize, std::string const & outFileName, bool isSingle) {


	_RTime=time(0);
	_CPUTime=clock();

	_maxTime=0;
	_timeStampMin=0;
	_timeStampMax=0;
	_isSingle=isSingle;

	_energy=new Int_t[_maxEvent];
	_adc=new Short_t[_maxEvent];
	_timeStamp=new Short_t[_maxEvent];
	_isClean=new Bool_t[_maxEvent];

	//_multDistrib=new Int_t[_maxEvent];

	//coinc root tree
	string rootFileName= outFileName + ".root";
	_rootFile = new TFile(const_cast<char*>(rootFileName.c_str()), "RECREATE");

	string TreeName = "Lst2RootTree";
	_rootTree = new TTree(const_cast<char*>(TreeName.c_str()),
			const_cast<char*>(TreeName.c_str()));

	_rootTree->Branch("multiplicity", &_multiplicity,"multiplicity/I");
	_rootTree->Branch("energy", _energy, "energy[multiplicity]/I");
	_rootTree->Branch("adc", _adc, "adc[multiplicity]/S");
	_rootTree->Branch("timeStamp", _timeStamp, "timeStamp[multiplicity]/S");
	_rootTree->Branch("isClean", _isClean, "isClean[multiplicity]/O");

	//single root Tree
	if (_isSingle){
		string rootFileSingleName= outFileName + "_Single" + ".root";
		_rootFileSingle = new TFile(const_cast<char*>(rootFileSingleName.c_str()), "RECREATE");

		string TreeNameSingle = "SingleTree";
		_rootTreeSingle = new TTree(const_cast<char*>(TreeNameSingle.c_str()),
			const_cast<char*>(TreeNameSingle.c_str()));

		_eventTreeSingle=new Event();

		_rootTreeSingle->Branch("Time", new long(_eventTreeSingle->time()), "timeStamp/L");
		//_rootTreeSingle->Branch("crate", &_eventTreeSingle->crate, "crate/S");
		_rootTreeSingle->Branch("adc", &_eventTreeSingle->channel, "adc/S");
		_rootTreeSingle->Branch("energy", new uint16_t(eventEnergy(*_eventTreeSingle)), "energy/I");
	}

	// organize detector IDs
	int16_t detectorIDMax = 0;
	for (vector<int16_t>::const_iterator d = detectorADCs.begin(); d != detectorADCs.end(); d++) {
		if (*d < 0) {
			int16_t detectorID = -1 - (*d);
			_detectorChannels.push_back(detectorID);
			_antiComptonChannels.push_back(true);

			if (detectorID > detectorIDMax) {
				detectorIDMax = detectorID;
			}

		} else {
			int16_t detectorID = *d;
			_detectorChannels.push_back(detectorID);
			_antiComptonChannels.push_back(false);

			if (detectorID > detectorIDMax) {
				detectorIDMax = detectorID;
			}
		}
	}

	_lastAntiComptonTimeADCs.resize(detectorIDMax + 1);

	// Generate the channel offset and the total size of the histograms
	BoardIterator b;
	int32_t currentNbChannels = (*b).nbChannels;

	++b;
	while (!b.end()) {
		_channelOffset[b.crate()][b.board()] = currentNbChannels;
		cout << "channel offset " << (int)b.crate() << ", " << (int)b.board() << " = " << _channelOffset[b.crate()][b.board()] << endl;
		currentNbChannels += (*b).nbChannels;

		++b;
	}

	// The total number of channels is the current number of channels
	_totalNbChannels = currentNbChannels;

	// Allocate the channel counters and rates
	_channelCounters = new uint64_t[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelCounters[i] = 0;
	}

	_channelRates = new double[_totalNbChannels];
	for (int32_t i = 0; i < _totalNbChannels; ++i) {
		_channelRates[i] = 0.0;
	}

	_maxNumberOfEnergyChannels = (1 << MAX_BITS);
	_coincidenceDT = coincidenceDT;
	_cleanDT = cleanDT;
	_minMult=mult;

	_numberOfGe=0;
	_numberOfBGO=0;

	_runQA = new QA();

	// TODO number of energy channels depends on the board
	_runQA->Init(_totalNbChannels, _maxNumberOfEnergyChannels, listModeContext.nbEvents, _maxEvent, _energyMax, _maxTimeCor, _matrixSize);


	_numberOfGoodEvents = 0;
	_numberOfBadEvents = 0;
	_numberOfEventsNotInserted = 0;
	_numberOfTruncatedCoincidences = 0;

	_blockSize=maxBlockSize;
}

CoincidenceOffline::~CoincidenceOffline() {
	delete _runQA;
	delete[] _energy;
	delete[] _adc;
	delete[] _timeStamp;
	delete[] _isClean;
	delete[] _channelCounters;
	if (_isSingle){
		delete _eventTreeSingle;
	}
}

int32_t CoincidenceOffline::numberOfGoodEvents() const {
	return _numberOfGoodEvents;
}

int32_t CoincidenceOffline::numberOfBadEvents() const {
	return _numberOfBadEvents;
}

int32_t CoincidenceOffline::numberOfEventsNotInserted() const {
	return _numberOfEventsNotInserted;
}

int32_t CoincidenceOffline::numberOfTruncatedCoincidences() const {
	return _numberOfTruncatedCoincidences;
}

bool CoincidenceOffline::checkEvent(Event const & event) {

	if (event.crate >= listModeContext.crateBoard.nbCrates) {
		cout << "WARNING: data probably corrupted, event " << event << " is bad, crate error" << endl;
		return false;
	}

	if (event.board >= listModeContext.crateBoard.crates[event.crate].nbBoards) {
		cout << "WARNING: data probably corrupted, event " << event << " is bad, board error" << endl;
		return false;
	}

	if (event.channel >= listModeContext.crateBoard.crates[event.crate].boards[event.board].nbChannels) {
		cout << "WARNING: data probably corrupted, event " << event << " is bad, channel error" << endl;
		return false;
	}

    return true;
}

void CoincidenceOffline::processBadEvent(Event const & event) {
	_numberOfBadEvents++;
}

void CoincidenceOffline::processEvent(Event & event) {
	_numberOfGoodEvents++;

	// updating ADC Counter
	int32_t index = _channelOffset[event.crate][event.board] + event.channel;

	_channelCounters[index]++;
	_channelRates[index] = pow(10, listModeContext.timeBase) * (double)_channelCounters[index] / (double)event.time();

	// Root
	_runQA->_htimeVsEvt->SetBinContent(_numberOfGoodEvents,event.time());

	_runQA->_hADCSpectra[index]->Fill(eventEnergy(event));
	Double_t E=eventEnergy(event) + (double)rand() / (double)RAND_MAX;
	_runQA->_hADCSpectraCal[index]->Fill(_detectors[index].Cal.first*E+_detectors[index].Cal.second);

	if (_isSingle){
		*_eventTreeSingle=event;
		//cout<<"energy: "<<_eventTreeSingle->energy<<endl;
		//cout<<"energy2: "<<_eventTreeSingle->gate<<endl;
		_rootTreeSingle->Fill();
		//cout<<" energy: "<<_eventTreeSingle->energy<<endl;
		//cout<<" energy2: "<<_eventTreeSingle->gate<<endl;
	}

	event.setTime(event.time()+_detectors[index].timeShift); //A VERIF SHIFT

	if (event.time()<_timeStampMin) _timeStampMin=event.time();
	if (event.time()>_timeStampMax) _timeStampMax=event.time();

}

void CoincidenceOffline::processEvent(EventBlockArrayIterator const & e) {
	_numberOfGoodEvents++;
}

void CoincidenceOffline::processEventNotInserted(Event const & event) {
	_numberOfGoodEvents++;
	_numberOfEventsNotInserted++;
}

int16_t CoincidenceOffline::detector(Event const & event) {
	int32_t index = _channelOffset[event.crate][event.board] + event.channel;
	return _detectorChannels[index];
}

bool CoincidenceOffline::antiCompton(Event const & event) {
	int32_t index = _channelOffset[event.crate][event.board] + event.channel;
	return _antiComptonChannels[index];
}

void CoincidenceOffline::setLastAntiComptonTime(Event const & event) {
	_lastAntiComptonTimeADCs[detector(event)] = event.time();
}

time_type CoincidenceOffline::lastAntiComptonTime(Event const & event) {
	return _lastAntiComptonTimeADCs[detector(event)];
}

int CoincidenceOffline::readConfigFile(string const & XPConfigFileName) {

	cout << "Reading configuration file '" << XPConfigFileName << "' ..."
			<< endl;

	ifstream file(XPConfigFileName.c_str());
	string line;
	int nbADCs=0;

	while (getline(file, line)) {
		if (line.compare(0, 1, "#") == 0)
			continue; //skip comments
		nbADCs++;

		Detector d;
		int trig;

		istringstream iss(line);
		iss >> d.channel >> d.detType.first;
		if (d.detType.first==1) _numberOfGe++;
		if (d.detType.first==-1) _numberOfBGO++;
		iss >> trig;
		if (trig==1){
			d.isTrig=true;
		}else{
			d.isTrig=false;
		}
		iss >> d.Cal.first >> d.Cal.second >> d.Thr.first >> d.Thr.second >> d.timeShift;

		_detectors.push_back(d);
	}

	return nbADCs;
}

void CoincidenceOffline::finish() {

	cout<<endl<<"> Filling QA histograms..."<<endl;

	for (int iADC=0; iADC<_totalNbChannels; iADC++){
		if (TMath::Finite(_channelRates[iADC])!=0){
			_runQA->_hADCRates->Fill(iADC,_channelRates[iADC]);
		}else{
			_runQA->_hADCRates->Fill(iADC,0.);
		}
	}

	_runQA->FillRelativeTime();

	_runQA->_hEvents->SetBinContent(1,_numberOfGoodEvents);
	_runQA->_hEvents->SetBinContent(2,_numberOfEventsNotInserted);
	_runQA->_hEvents->SetBinContent(3,_numberOfTruncatedCoincidences);
	_runQA->_hEvents->SetBinContent(4,_blockSize);

	cout<<endl<<"> Writing tree and histograms..."<<endl;

	_rootFile->cd();
	_rootTree->Write();
	TDirectory *dirQA=_rootFile->mkdir("QA");
	TDirectory *dirADCS=dirQA->mkdir("ADCSpectraNoCal");
	TDirectory *dirADCSC=dirQA->mkdir("ADCSpectraCal");
	TDirectory *dirADCSCC=dirQA->mkdir("ADCSpectraCalClean");

	_rootFile->cd("QA");
	_runQA->_hMult->Write();
	_runQA->_hADCRates->Write();
	_runQA->_htimeVsEvt->Write();
	_runQA->_hRelativeTimeTH2->Write();
	_runQA->_hCoincMatrix->Write();
	_runQA->_hEvents->Write();

	_runQA->_hRelativeTimeTH1[8]->Write();
	_runQA->_hRelativeTimeTH1[128]->Write();

	_rootFile->cd("QA/ADCSpectraNoCal");
	for (int iadc=0; iadc<_totalNbChannels;iadc++){
		_runQA->_hADCSpectra[iadc]->Write();
	}
	_rootFile->cd("QA/ADCSpectraCal");
	for (int iadc=0; iadc<_totalNbChannels;iadc++){
		_runQA->_hADCSpectraCal[iadc]->Write();
	}
	_rootFile->cd("QA/ADCSpectraCalClean");
	for (int iadc=0; iadc<_totalNbChannels;iadc++){
		_runQA->_hADCSpectraCalClean[iadc]->Write();
	}

	_rootFile->cd("QA");
	_RTime=(time(0)-_RTime);
	_CPUTime=(clock()-_CPUTime)/(CLOCKS_PER_SEC);
	_runQA->_hSyst->SetBinContent(1,_RTime);
	_runQA->_hSyst->SetBinContent(2,_CPUTime);
	_runQA->_hSyst->Write();

	_rootFile->Close();

	if (_isSingle){
		_rootFileSingle->cd();
		_rootTreeSingle->Write();
		_rootFileSingle->Close();
	}

	cout<<endl;
	cout<<"> Acquisition time: "<<_timeStampMax-_timeStampMin<<" TS"<<endl;
	cout<<"      with TS=10ns: "<<(_timeStampMax-_timeStampMin)*10*pow(10.,-9) <<" s"<<endl;
	cout<<endl;

}


//######################################################################
//AJOUT LST2ROOT


bool CoincidenceOffline::isNoise(Event const & event){
	//return 'true' if the event is out of the adc channel range [rangeMin, rangeMax], 'false' else
	bool noise=true;
	int16_t index = _channelOffset[event.crate][event.board] + event.channel;

	if ((_detectors[index].Thr.first<=eventEnergy(event))&&(eventEnergy(event)<=_detectors[index].Thr.second)) noise=false;

	return noise;
}

bool CoincidenceOffline::isTrig(Event const & event){
	bool trig=false;
	int16_t index = _channelOffset[event.crate][event.board] + event.channel;
	if (!isNoise(event)&&(_detectors[index].isTrig)) trig=true;

	return trig;
}

bool CoincidenceOffline::isToBeFilled(Event const & event){
	bool toBeFilled=false;
	int16_t index = _channelOffset[event.crate][event.board] + event.channel;
	if (_detectors[index].detType.first==1) toBeFilled=true;

	return toBeFilled;

}

int32_t CoincidenceOffline::getTotalNumberOfChannels() const {
	return _totalNbChannels;
}

double * CoincidenceOffline::getChannelRates() const {
	return _channelRates;
}

void CoincidenceOffline::printRates() {

	cout << "adc rates" << endl;

	BoardIterator b;
	int32_t channel = 0;
	while (!b.end()) {

		cout << "crate " << (int)b.crate() << " board " << (int)b.board() << endl;
		for (int32_t c = 0; c < (*b).nbChannels; ++c) {
			cout << "\t" << setw(5) << c << setw(10) << _channelRates[channel] << endl;
			++channel;
		}
		cout << endl;

		++b;
	}
}

}
