/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LSTDPP_COINCIDENCEOFFLINE_H
#define LSTDPP_COINCIDENCEOFFLINE_H

#include <cstring>
#include <algorithm>
#include <fstream>
#include <vector>
#include <time.h>
#include <sstream>

#include <TTree.h>
#include <TFile.h>
#include <TH1.h>
#include <TH2.h>
#include <TMath.h>
#include <TDirectory.h>

#include "../../Event.h"
#include "../../EventData.h"
#include "../../EventBlockArray.h"
#include "Detector.h"
#include "QA.h"

namespace lstdpp128 {

class CoincidenceOffline {

public:
	CoincidenceOffline();
	~CoincidenceOffline();

	void init(std::vector<int16_t> const & detectorADCs, time_type coincidenceDT, time_type cleanDT, int mult, int32_t maxBlockSize, std::string const & outFileName, bool isSingle); //AJOUT LST2ROOT

	int32_t numberOfGoodEvents() const;
	int32_t numberOfBadEvents() const;
	int32_t numberOfEventsNotInserted() const;
	int32_t numberOfTruncatedCoincidences() const;
	int16_t detector(Event const & event);
	bool antiCompton(Event const & event);

	/**
	 * Functions to implement.
	 */
	bool checkEvent(Event const & event);
	void processBadEvent(Event const & event);
	void processEvent(Event & event);
	void processEvent(EventBlockArrayIterator const & e);
	void processEventNotInserted(Event const & event);
	int readConfigFile(std::string const & XPConfigFileName);

	template<typename Iterator>
	Iterator processBlock(Iterator first, Iterator last, bool lastBlock);
	void finish();
	void printRates();

	int32_t getTotalNumberOfChannels() const;
	double * getChannelRates() const;

private:

	void setLastAntiComptonTime(Event const & event);
	time_type lastAntiComptonTime(Event const & event);

	template<typename Iterator>
	Iterator calculateCleanEvents(Iterator first, Iterator last, bool lastBlock);

	template<typename Iterator>
	Iterator calculateCoincidences(Iterator first, Iterator last, bool lastBlock);

	bool isNoise(Event const & event);
	bool isTrig(Event const & event);
	bool isToBeFilled(Event const & event);
	template<typename Iterator>
	std::pair<std::vector<Event>,Iterator>  eventsInCoincArray(Iterator first, Iterator last, double deltaT);
	template<typename Iterator>
	void fillTree(std::pair<std::vector<Event>, Iterator> evtInCoinc);

	// crate, board - channel offset
	std::map<int32_t, std::map<int32_t, int32_t> > _channelOffset;

	std::vector<int16_t> _detectorChannels;
	std::vector<bool> _antiComptonChannels;
	std::vector<time_type> _lastAntiComptonTimeADCs;
	int32_t _totalNbChannels;
	uint64_t * _channelCounters;
	double * _channelRates;
	int32_t _maxNumberOfEnergyChannels;
	time_type _coincidenceDT;
	time_type _cleanDT;
	int _minMult;

	int32_t _numberOfGoodEvents;
	int32_t _numberOfBadEvents;
	int32_t _numberOfEventsNotInserted;
	time_type _maxTime;
	int64_t _timeStampMin;
	int64_t _timeStampMax;

	int32_t _numberOfGe;
	int32_t _numberOfBGO;

	static const int _maxEvent=10000;
	static const int _matrixSize=4096;
	static const double _energyMax; //energy max in keV for coinc. matrix and histos
	static const int _maxTimeCor=10000;
	int32_t _numberOfTruncatedCoincidences;
	TTree * _rootTree;
	TFile * _rootFile;
	Int_t _multiplicity;
	Int_t *_energy;
	Short_t *_adc;
	Short_t *_timeStamp;
	Bool_t *_isClean;
	std::vector<Detector> _detectors;
	QA* _runQA;

	time_t _RTime;
	clock_t _CPUTime;
	int _blockSize;

	//single tree
	bool _isSingle;
	TTree * _rootTreeSingle;
	TFile * _rootFileSingle;
	Event * _eventTreeSingle;

	static const int16_t MAX_BITS = 15;
	static const int16_t GLOBAL = 0;
	static const int16_t GLOBAL_CLEAN = 1;
	static const int16_t DETECTOR = 2;
	static const int16_t DETECTOR_CLEAN = 3;
};

template<typename Iterator>
Iterator CoincidenceOffline::processBlock(Iterator first, Iterator last, bool lastBlock) {

	last = calculateCleanEvents(first, last, lastBlock);
	last = calculateCoincidences(first, last, lastBlock);
	//writeBlock(first, last); //A VERIF: POURQUOI WRITE??

	return last;
}

template<typename Iterator>
Iterator CoincidenceOffline::calculateCleanEvents(Iterator first, Iterator last, bool lastBlock) {

	Iterator e = first;
	Iterator processedLast = first;

	// iterating events
	while (e != last) {

		// forcing const_cast because iterators on multiset return const references
		Event & event = const_cast<Event &>(*e);
		int16_t index = _channelOffset[event.crate][event.board] + event.channel;
		int16_t detectorID = detector(event);

		if (antiCompton(event)) {
			setLastAntiComptonTime(event);
			event.status = Event::ANTI_COMPTON;
		} else {
			// change dT into special dT for anti-compton test
			if (event.time() - lastAntiComptonTime(event) <= _cleanDT) {
				event.status = Event::NOT_CLEAN;
			} else {
				Iterator b = e;
				++b;
				bool foundStatus = false;
				while (b != last) {
					if ((*b).time() - event.time() > _cleanDT) {
						event.status = Event::CLEAN;
						foundStatus = true;
						break;
					} else if (antiCompton(*b) && detector(*b) == detectorID) {
						event.status = Event::NOT_CLEAN;
						foundStatus = true;
						break;
					}

					++b;
				}

				if (!foundStatus && b == last) {
					processedLast = e;
					break;
				}
			}
		}


		if ((event.status==Event::CLEAN)&&(_detectors[index].detType.first==1)){
			Double_t E=eventEnergy(event) + (double)rand() / (double)RAND_MAX;
			_runQA->_hADCSpectraCalClean[index]->Fill(_detectors[index].Cal.first*E+_detectors[index].Cal.second);
		}

		//filling relative time distributions
		if (e!=first){
			Event & eventBe = const_cast<Event &>(*(e-1));
			int16_t indexBe = _channelOffset[eventBe.crate][eventBe.board] + eventBe.channel;
			if((-1000<=(event.time()-eventBe.time()))&&((event.time()-eventBe.time())<=_maxTimeCor)){
				_runQA->_hRelativeTimeTH1[index*_totalNbChannels+indexBe]->Fill(event.time()-eventBe.time());
			}
		}
		if (e!=last){
			Event & eventAf = const_cast<Event &>(*(e+1));
			int16_t indexAf = _channelOffset[eventAf.crate][eventAf.board] + eventAf.channel;
			if((-1000<=(event.time()-eventAf.time()))&&((event.time()-eventAf.time())<=_maxTimeCor)){
				_runQA->_hRelativeTimeTH1[index*_totalNbChannels+indexAf]->Fill(event.time()-eventAf.time());
			}
		}

		++e;
	}

	if (!lastBlock && processedLast != first) {
		//std::cout << "distance with last = " << std::distance(processedLast, last) << std::endl;
		return processedLast;
	}

	// in case of last block, considering that last events are clean
	while (e != last) {
		// forcing const_cast because iterators on multiset return const references
		Event & event = const_cast<Event &>(*e);

		if (antiCompton(event)) {
			event.status = Event::ANTI_COMPTON;
		} else {
			event.status = Event::CLEAN;
		}

		++e;
	}

	return last;
}

template<typename Iterator>
Iterator CoincidenceOffline::calculateCoincidences(Iterator first, Iterator last, bool lastBlock) {

	bool isClean = true;
	bool isDetector = true;
	std::set<int16_t> detectors;
	Iterator processedLast = first;
	Iterator e = first;


	// iterating events
	while (e != last) {

		Event const & event = *e;

		if (event.time()>_maxTime) _maxTime=event.time();


		if (isTrig(event)){
			Event evtTrig=event;
			std::pair<std::vector<Event>, Iterator> evtInCoinc;
			evtInCoinc=eventsInCoincArray(e,last,_coincidenceDT);
			processedLast=evtInCoinc.second;
			e=processedLast;
			if (evtInCoinc.first.size()>0){ //if 0 dummy event
				fillTree(evtInCoinc);
			}
		}else{
			processedLast=e;
		}

		++e;

	}

	if (!lastBlock && processedLast != first) {
		return processedLast;
	}

	if (!lastBlock) {
		_numberOfTruncatedCoincidences++;
	}

	// returning last in case of first coincidence
	return last;
}

template<typename Iterator>
std::pair<std::vector<Event>,Iterator> CoincidenceOffline::eventsInCoincArray(Iterator first, Iterator last, double deltaT) {

	Iterator e=first;
	int nbTrig=1;
	int nbTrigSup=0;
	std::pair<std::vector<Event>,Iterator> tempToReturn;
	std::vector<std::pair<std::vector<Event>,Iterator> > evtInCoinc;

	Event const & evtTrig = *e;
	tempToReturn.first.push_back(evtTrig);
	tempToReturn.second=e;
	evtInCoinc.push_back(tempToReturn);

	while (e!=last){
		Event const & evtCurrent = *e;

		if(e==first){
			//if e==first: evtCurrent is the trigger -> useless to process
			//!!!!!!!! keep inside the loop otherwise e!=last can fail => infinite loop
			//to be added: perform this check outside the eventsInCoincArray function??
			e++;
			continue;
		}

		if (evtCurrent.time()>_maxTime) _maxTime=evtCurrent.time();

		if (isNoise(evtCurrent)){
			e++;
			continue;
		}

		if (evtCurrent.time() > evtTrig.time() + 2*deltaT) break;

		for (int iTrig=nbTrigSup; iTrig<nbTrig; iTrig++){
			if (evtCurrent.time() <= evtInCoinc[iTrig].first[0].time()+deltaT){
				if (isToBeFilled(evtCurrent)){
					std::pair<std::vector<Event>,Iterator> temp;
					temp=evtInCoinc[iTrig];
					temp.first.push_back(evtCurrent);
					temp.second=e;
					evtInCoinc[iTrig]=temp;
				}
			}else{
				nbTrigSup++;
			}
		}

		if ((evtCurrent.time() <= evtTrig.time() + deltaT)&&(isTrig(evtCurrent))){
			std::pair<std::vector<Event>,Iterator> temp;
			temp.first.push_back(evtCurrent);
			temp.second=e;
			evtInCoinc.push_back(temp);
			nbTrig++;
		}

		e++;
	}

	if (e==last){ //if e==last -> end of block reached, to be reprocessed with the next block
		tempToReturn.first.clear();
		tempToReturn.second=e-1;
	}else{ //return the coincidence with the highest multiplicity
		tempToReturn=evtInCoinc[0];
		for (int iTrig=1; iTrig<nbTrig; iTrig++){
			if (evtInCoinc[iTrig].first.size() > tempToReturn.first.size()){
				tempToReturn=evtInCoinc[iTrig];
			}
		}
	}

	return tempToReturn;

}

template<typename Iterator>
void CoincidenceOffline::fillTree(std::pair<std::vector<Event>, Iterator> evtInCoinc){

	_multiplicity=evtInCoinc.first.size();
	if (_multiplicity>_maxEvent){
		_multiplicity=_maxEvent;
		std::cout<<"WARNING: maximum number of events in coincidence reached"<<std::endl;
		std::cout<<"         multiplicity "<<_maxEvent<<" forced in root Tree"<<std::endl<<std::endl;
	}
	for (int i=0; i<_multiplicity;i++){

		int16_t index = _channelOffset[evtInCoinc.first[i].crate][evtInCoinc.first[i].board] + evtInCoinc.first[i].channel;

		_energy[i]=eventEnergy(evtInCoinc.first[i]);
		_adc[i]=index;
		_timeStamp[i]=evtInCoinc.first[i].time()-evtInCoinc.first[0].time();

		if(evtInCoinc.first[i].status==Event::UNDEFINED){
			std::cout<<"PROBLEM: UNDEFINED status"<<std::endl;
		}

		if ((evtInCoinc.first[i].status==Event::CLEAN)||(evtInCoinc.first[i].status==Event::UNDEFINED)){
			_isClean[i]=true;
		}else{
			_isClean[i]=false;
		}
	}
	if (_multiplicity>=_minMult) _rootTree->Fill();

	_runQA->_hMult->Fill(_multiplicity);

	if (_multiplicity>=2){
		for (int i1=0; i1<evtInCoinc.first.size(); i1++){

			int16_t index1 = _channelOffset[evtInCoinc.first[i1].crate][evtInCoinc.first[i1].board] + evtInCoinc.first[i1].channel;

			if (_detectors[index1].detType.first==1) {
				Double_t E1=_energy[i1] + (double)rand() / (double)RAND_MAX;
				Double_t ECal1=_detectors[index1].Cal.first*E1+_detectors[index1].Cal.second;

				for (int i2=i1+1; i2<evtInCoinc.first.size(); i2++){

					int16_t index2 = _channelOffset[evtInCoinc.first[i2].crate][evtInCoinc.first[i2].board] + evtInCoinc.first[i2].channel;

					if (_detectors[index2].detType.first==1) {
						Double_t E2=_energy[i2] + (double)rand() / (double)RAND_MAX;
						Double_t ECal2=_detectors[index2].Cal.first*E2+_detectors[index2].Cal.second;

						int bin1= _runQA->_hCoincMatrix->FindBin(ECal1);
						int bin2= _runQA->_hCoincMatrix->FindBin(ECal2);
						_runQA->_hCoincMatrix->Fill(ECal1, ECal2);
						if (bin1!=bin2) _runQA->_hCoincMatrix->Fill(ECal2, ECal1);
					}
				}
			}
		}
	}
}

}

#endif
