/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <iostream>
#include <stdint.h>
#include <fstream>
#include <cstdlib>
#include <cstdio>
#include <iomanip>

#include "../../EventData.h"
#include "../../Reader.h"

using namespace std;
using namespace lstdpp128;

const int BUFFER_SIZE = 1 << 10;

std::vector<int32_t> split(const std::string& content) {

	vector<int32_t> result;

	int lastIndex = 0;
	int index = content.find(',');
	while (index != string::npos) {

		istringstream is(content.substr(lastIndex, index - lastIndex));
		int32_t value;
		is >> value;

		result.push_back(value);
		lastIndex = index + 1;
		index = content.find(',', lastIndex);
	}

	istringstream is(content.substr(lastIndex));
	int32_t value;
	is >> value;

	result.push_back(value);

	return result;
}


std::string eventColumnsToCSV() {

	ostringstream os;

	os << "time;"
		<< "rollover;"
		<< "timestamp;"
		<< "crate;"
		<< "board;"
		<< "channel;"
		<< "board type;";

	return os.str();
}

std::string eventDataColumnsToCSV(BoardType type) {

	ostringstream os;

	// data[0]
	if (boardHasEnergy(type)) {
		os << "energy;";
	}

	if (boardHasQ(type)) {
		os << "q-short;";
		os << "q-long;";
	}

	if (boardHasAB(type)) {
		os << "A;";
		os << "A+B;";
	}

	if (boardHasPositionX(type)) {
		os << "X;";
	}

	if (boardHasPositionY(type)) {
		os << "Y;";
	}

	// data[1]
	if (boardHasZeroCrossing(type)) {
		os << "0-crossing;";
	}

	if (boardHasUnspecifiedData(type)) {
		os << "unspec-data;";
	}

	// flags.
	if (boardHasFlags(type)) {
		os << "flags;";
	}

	return os.str();
}

std::string eventToCSV(Event const & event) {

	ostringstream os;

	os << event.time() << ";"
		<< event.rollover << ";"
		<< event.timestamp << ";"
		<< (int)event.crate << ";"
		<< (int)event.board << ";"
		<< event.channel << ";"
		<< (int)event.boardType() << ";";

	return os.str();
}

std::string eventDataToCSV(Event const & event) {

	ostringstream os;

	// data[0]
	if (eventHasEnergy(event)) {
		os << eventEnergy(event) << ";";
	}

	if (eventHasQ(event)) {
		os << eventQShort(event) << ";";
		os << eventQLong(event) << ";";
	}

	if (eventHasAB(event)) {
		os << eventA(event) << ";";
		os << eventAplusB(event) << ";";
	}

	if (eventHasPositionX(event)) {
		os << eventPositionX(event) << ";";
	}

	if (eventHasPositionY(event)) {
		os << eventPositionY(event) << ";";
	}

	// data[1]
	if (eventHasZeroCrossing(event)) {
		os << eventZeroCrossing(event) << ";";
	}

	// special
	if (eventHasUnspecifiedData(event)) {
		os << eventUnspecifiedData(event) << ";";
	}

	// flags
	if (eventHasFlags(event) != 0) {

		ostringstream bos;

		uint8_t flags = eventFlags(event);
		for (int i = 7; i >= 0; --i) {
			uint8_t flag = (flags >> i) & 0x1;
			if (flag == 1) {
				bos << "1";
			}
			else {
				bos << "0";
			}
		}

		os << bos.str() << ";";
	}

	return os.str();
}

void process(int32_t * buffer, int32_t size, int32_t crate, int32_t board, int32_t channel, int numberOfEvents, int32_t& eventId, ofstream & outfile) {

	// Reading next block.
	int32_t index = 0;
	int32_t nbIndex = size;
	int32_t * bufferPtr = buffer;

	while (index < nbIndex) {
		// Reading event.
		Event event;
		if (readEvent(event, bufferPtr + index)) {

			bool display = ((crate == -1 || (crate == event.crate))
						&& (board == -1 || (board == event.board))
						&& (channel == -1 || (channel == event.channel)));

			if (display) {
				outfile << eventToCSV(event) << eventDataToCSV(event) << endl;
			}

		} else {
			cout << "problem with event " << hex << "    " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index))
										<< " " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index + 1))
										<< " " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index + 2))
										<< " " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index + 3))
										<< setw(0) << setfill(' ') << dec << endl;
		}

		index += 4;
		eventId++;

		// Break if number of events is reached.
		if (numberOfEvents != -1 && eventId >= numberOfEvents) {
			cout << "exiting " << eventId << endl;
			break;
		}
	}
}

int main(int argc, char * argv[]) {

	if (argc < 3) {
		cerr << "usage : lstreader128 <file> <outfile> [filter] [number of events]" << endl;
		cerr << "  filter is [crate,board,channel] or [crate,board] or [crate]" << endl;
		return EXIT_FAILURE;
	}

	string fileName = argv[1];

	int32_t crate = -1;
	int32_t board = -1;
	int32_t channel = -1;

	string outfileName = argv[2];

	if (argc >= 4) {
		string valuesString(argv[3]);
		valuesString = valuesString.substr(1, valuesString.length() - 2);
		vector<int32_t> values = split(valuesString);

		cout << "filter:" << endl;

		if (values.size() > 0) {
			crate = values[0];
			cout << "crate   " << crate << endl;
		}

		if (values.size() > 1) {
			board = values[1];
			cout << "board   " << board << endl;
		}

		if (values.size() > 2) {
			channel = values[2];
			cout << "channel " << channel << endl;
		}
	}

	int numberOfEvents = -1;
	if (argc >= 5) {
		istringstream is(argv[4]);
		is >> numberOfEvents;
	}

	cout << "number of events is " << numberOfEvents << endl;

	Reader reader(BUFFER_SIZE);

	bool success = reader.open(fileName);

	if (!success) {
		return EXIT_FAILURE;
	}

	cout << listModeContext << endl;

	cout << "flags are:" << endl;
	cout << "00100000: pile-up" << endl;
	cout << "00010000: dead time/lost events" << endl;
	cout << "00001000: over-range" << endl;
	cout << "00000100: trigger lost" << endl;
	cout << "00000010: n trigger counted" << endl;
	cout << "00000001: n lost trigger counted" << endl << endl;

	// Print the columns of the event data.
	BoardIterator b;
	set<BoardType> boardTypeSet;
	while (!b.end()) {
		boardTypeSet.insert((*b).boardType);
		++b;
	}

	cout << setw(10) << "board type" << endl;

	for (set<BoardType>::const_iterator t = boardTypeSet.begin(); t != boardTypeSet.end(); ++t) {
		cout << "type " << *t << ":" << endl;
		cout << eventColumnsToCSV() << eventDataColumnsToCSV(*t) << endl << endl;
	}

	int32_t eventId = 0;

	// Open the outfile.
	ofstream outfile;
	outfile.open(outfileName.c_str());

	// Reading.
	while (true) {
		// Reading next block.
		int32_t size = reader.read();

		if (size == 0) {
			break;
		}

		// Processing.
		process(reader.buffer(), size, crate, board, channel, numberOfEvents, eventId, outfile);

		// Break if number of events is reached.
		if (numberOfEvents != -1 && eventId >= numberOfEvents) {
			cout << "exiting " << eventId << endl;
			break;
		}
	}

	// Close the file.
	outfile.close();

	return EXIT_SUCCESS;
}
