/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <iostream>
#include <stdint.h>
#include <fstream>
#include <cstdlib>
#include <cstdio>
#include <iomanip>

#include "../../EventData.h"
#include "../../Reader.h"

using namespace std;
using namespace lstdpp128;

const int BUFFER_SIZE = 1 << 10;

std::vector<int32_t> split(const std::string& content) {

	vector<int32_t> result;

	int lastIndex = 0;
	int index = content.find(',');
	while (index != string::npos) {

		istringstream is(content.substr(lastIndex, index - lastIndex));
		int32_t value;
		is >> value;

		result.push_back(value);
		lastIndex = index + 1;
		index = content.find(',', lastIndex);
	}

	istringstream is(content.substr(lastIndex));
	int32_t value;
	is >> value;

	result.push_back(value);

	return result;
}

void process(int32_t * buffer, int32_t size, int32_t crate, int32_t board, int32_t channel, int32_t group, int32_t& eventId) {

	// reading next block
	int32_t index = 0;
	int32_t nbIndex = size;
	int32_t * bufferPtr = buffer;

	while (index < nbIndex) {
		// reading event
		Event event;
		if (readEvent(event, bufferPtr + index)) {

			bool display = ((crate == -1 || (crate == event.crate))
						&& (board == -1 || (board == event.board))
						&& (channel == -1 || (channel == event.channel)));

			if (display) {
				cout << event << eventDataToString(event) << " |" << eventId << "| " << printBinary(bufferPtr + index) << endl;

				if ((eventId + 1) % group == 0) {
					cout << endl;
				}
			}

		} else {
			cout << "problem with event " << hex << "    " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index))
										<< " " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index + 1))
										<< " " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index + 2))
										<< " " << setw(8) << setfill('0') << (uint32_t)(*(bufferPtr + index + 3))
										<< setw(0) << setfill(' ') << dec << endl;
		}

		index += 4;
		eventId++;
	}
}

int main(int argc, char * argv[]) {

	if (argc < 2) {
		cerr << "usage : lstreader128 <file> [group] [filter]" << endl;
		cerr << "  filter is [crate,board,channel] or [crate,board] or [crate]" << endl;
		return EXIT_FAILURE;
	}

	string fileName = argv[1];

	int32_t crate = -1;
	int32_t board = -1;
	int32_t channel = -1;

	int32_t group = 4096;
	if (argc >= 3) {
		string groupString(argv[2]);
		istringstream is(groupString);
		is >> group;
	}

	cout << "grouping by " << group << endl;

	if (argc >= 4) {
		string valuesString(argv[3]);
		valuesString = valuesString.substr(1, valuesString.length() - 2);
		vector<int32_t> values = split(valuesString);

		cout << "filter:" << endl;

		if (values.size() > 0) {
			crate = values[0];
			cout << "crate   " << crate << endl;
		}

		if (values.size() > 1) {
			board = values[1];
			cout << "board   " << board << endl;
		}

		if (values.size() > 2) {
			channel = values[2];
			cout << "channel " << channel << endl;
		}
	}

	Reader reader(BUFFER_SIZE);

	bool success = reader.open(fileName);

	if (!success) {
		return EXIT_FAILURE;
	}

	cout << listModeContext << endl;

	// Print the columns of the event data.
	BoardIterator b;
	set<BoardType> boardTypeSet;
	while (!b.end()) {
		boardTypeSet.insert((*b).boardType);
		++b;
	}

	cout << setw(10) << "board type" << endl;

	for (set<BoardType>::const_iterator t = boardTypeSet.begin(); t != boardTypeSet.end(); ++t) {
		cout << eventColumnsToString(*t) << eventDataColumnsToString(*t) << endl;
	}

	int32_t eventId = 0;

	// reading
	while (true) {
		// reading next block
		int32_t size = reader.read();

		if (size == 0) {
			break;
		}

		// processing
		process(reader.buffer(), size, crate, board, channel, group, eventId);
	}

	return EXIT_SUCCESS;
}
