/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <cstdlib>
#include <fstream>
#include <sstream>
#include <cstring>
#include <algorithm>
#include <time.h>

#include "../../Check.h"
#include "../../Event.h"
#include "../../EventArrayAlgorithm.h"
#include "../../EventBlockArray.h"
#include "../../Reader.h"

using namespace std;
using namespace lstdpp128;

int32_t BUFFER_SIZE = 1 << 12;
int32_t BLOCK_SIZE = 1 << 13;

int percentage(int a, int b) {
	return 100.0f * (float)a / (float)b;
}

void printRate(double * rate) {

	for (int i = 0; i < 7; ++i) {
		cout << rate[i] << " ";
	}
	cout << endl;
}

int main(int argc, char * argv[]) {

	if (argc < 2) {
		cerr << "usage : lstcheck128 <file> [blockSize]" << endl;
		return EXIT_FAILURE;
	}

	string fileName = argv[1];

	int32_t blockSize = BLOCK_SIZE;

	if (argc >= 3) {
		istringstream im(argv[2]);
		im >> blockSize;
	}

	cout << "reading " << fileName << ", block size " << blockSize << endl;

	Reader reader(BUFFER_SIZE);

	bool success = reader.open(fileName);

	if (!success) {
		return EXIT_FAILURE;
	}

	// algorithm
	EventArrayAlgorithm<Check> algorithm;
	algorithm.processor().init();

	// creating the events
	EventArray events(blockSize);

	int currentPercentage = 0;

	time_t start = time(0);
	time_t t1 = start;

	// reading list mode
	while (true) {
		// reading next block
		int32_t size = reader.read();

		// exiting if no data
		if (size == 0) {
			break;
		}

		// processing block
		algorithm.process(reader.buffer(), size, events);
	}

	algorithm.finish(events);

	time_t t2 = time(0);

	cout << "terminated in " << (t2 - t1) << "s" << endl;

	cout << "#events read            = " << algorithm.processor().numberOfGoodEvents() << " (" << ((float)algorithm.processor().numberOfGoodEvents() / (float)(t2 - t1)) << " event/s)" << endl;
	cout << "#events bad             = " << algorithm.processor().numberOfBadEvents() << " (" << percentage(algorithm.processor().numberOfBadEvents(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;
	cout << "#events not inserted    = " << algorithm.processor().numberOfEventsNotInserted() << " (" << percentage(algorithm.processor().numberOfEventsNotInserted(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;
	cout << "max time                = " << algorithm.processor().maxTime() << endl;
	if (algorithm.processor().isOrdered()) {
		cout << "ordered" << endl;
	} else {
		cout << "not ordered" << endl;
	}

	return EXIT_SUCCESS;
}
