/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <cstdlib>
#include <fstream>
#include <sstream>
#include <cstring>
#include <algorithm>
#include <time.h>

#include "../../algorithms/nomad/Histogram.h"
#include "../../Event.h"
#include "../../EventBlockArray.h"
#include "../../EventBlockArrayAlgorithm.h"
#include "../../Reader.h"

using namespace std;
using namespace lstdpp128;

int32_t MAX_BLOCK_SIZE = 0;
int32_t BUFFER_SIZE = 1 << 12;

int percentage(int a, int b) {
	return 100.0f * (float)a / (float)b;
}

int main(int argc, char * argv[]) {

	if (argc < 7) {
		cerr << "usage : lsthisto128 <file> <crate> <board> <channel> <delay> <dT> [maxBlocks] [blockSize]" << endl;
		return EXIT_FAILURE;
	}

	string fileName = argv[1];

	istringstream icc(argv[2]);
	uint32_t crate;
	icc >> crate;

	istringstream icb(argv[3]);
	uint32_t board;
	icb >> board;

	istringstream ich(argv[4]);
	uint32_t channel;
	ich >> channel;

	istringstream id(argv[5]);
	time_type delay;
	id >> delay;

	istringstream it(argv[6]);
	time_type dT;
	it >> dT;

	int32_t maxBlocks = -1;

	if (argc >= 8) {
		istringstream im(argv[7]);
		im >> maxBlocks;
	}

	if (argc >= 9) {
		istringstream is(argv[8]);
		is >> BUFFER_SIZE;
	}

	MAX_BLOCK_SIZE = BUFFER_SIZE * 2;

	cout << "reading " << fileName << ", channel " << channel << ", " << delay << ", dT " << dT << ", max blocks " << maxBlocks << ", block size " << BUFFER_SIZE << endl;
	string outFileName = fileName.substr(0, fileName.find_first_of('.')) + ".lsto";

	Reader reader(BUFFER_SIZE);

	bool success = reader.open(fileName);

	if (!success) {
		return EXIT_FAILURE;
	}

	// algorithm
	EventBlockArrayAlgorithm<Histogram> algorithm;
	algorithm.processor().init(reader.boardHeader(), reader.version(), reader.timeBase(), reader.totalNumberOfBoards(), crate, board, channel, delay, dT, MAX_BLOCK_SIZE, outFileName);

	// creating the events
	EventBlockArray blocks(MAX_BLOCK_SIZE, maxBlocks);

	int currentPercentage = 0;

	time_t start = time(0);
	time_t t1 = start;

	// reading list mode
	while (true) {
		// reading next block
		int32_t size = reader.read();

		// exiting if no data
		if (size == 0) {
			break;
		}

		// processing block
		algorithm.process(reader.buffer(), size, blocks);
	}

	time_t t2 = time(0);

	algorithm.finish(blocks);

	cout << "#events read         = " << algorithm.processor().numberOfGoodEvents() << " (" << ((float)algorithm.processor().numberOfGoodEvents() / (float)(t2 - t1)) << " event/s)" << endl;
	cout << "#events bad          = " << algorithm.processor().numberOfBadEvents() << " (" << percentage(algorithm.processor().numberOfBadEvents(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;
	cout << "#events not inserted = " << algorithm.processor().numberOfEventsNotInserted() << " (" << percentage(algorithm.processor().numberOfEventsNotInserted(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;
	cout << "#events in conflict  = " << algorithm.processor().numberOfEventsInConflict() << " (" << percentage(algorithm.processor().numberOfEventsInConflict(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;

	algorithm.processor().printRates();

	return EXIT_SUCCESS;
}
