/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <cstdlib>
#include <fstream>
#include <sstream>
#include <cstring>
#include <algorithm>
#include <time.h>

#include "../../algorithms/nomad/Coincidence.h"
#include "../../Event.h"
#include "../../EventArrayAlgorithm.h"
#include "../../EventBlockArray.h"
#include "../../Reader.h"

using namespace std;
using namespace lstdpp128;

int32_t BUFFER_SIZE = 1 << 12;

int percentage(int a, int b) {
	return 100.0f * (float)a / (float)b;
}

int main(int argc, char * argv[]) {

	if (argc < 5) {
		cerr << "usage : lstcoincidence128 <file> <coincidence dT> <cleanDT> <detectoradc.txt> <coincidenceMatrixResolution> [blockSize]" << endl;
		return EXIT_FAILURE;
	}

	string fileName = argv[1];

	istringstream cit(argv[2]);
	time_type coincidenceDT;
	cit >> coincidenceDT;

	istringstream clit(argv[3]);
	time_type cleanDT;
	clit >> cleanDT;

	string detectorADCFileName = argv[4];
	vector<int16_t> detectorADCList = loadList<int16_t>(detectorADCFileName);

	int16_t coincidenceMatrixResolution = 12;

	if (argc >= 6) {
		istringstream ir(argv[5]);
		ir >> coincidenceMatrixResolution;
	}

	int32_t blockSize = BUFFER_SIZE * 2;

	if (argc >= 7) {
		istringstream im(argv[6]);
		im >> blockSize;
	}

	cout << "reading " << fileName << ", coincidence dT " << coincidenceDT << ", clean dT " << cleanDT << ", block size " << blockSize << endl;
	cout << "detector ADC = " << printList<int16_t>(detectorADCList) << endl;

	string outFileName = fileName.substr(0, fileName.find_last_of('.')) + ".lsto";

	Reader reader(BUFFER_SIZE);

	bool success = reader.open(fileName);

	if (!success) {
		return EXIT_FAILURE;
	}

	// algorithm
	EventArrayAlgorithm<Coincidence> algorithm;
	algorithm.processor().init(reader.boardHeader(), reader.version(), reader.timeBase(), reader.totalNumberOfBoards(), coincidenceMatrixResolution, detectorADCList, coincidenceDT, cleanDT, blockSize, outFileName);

	// creating the events
	EventArray events(blockSize);

	int currentPercentage = 0;

	time_t start = time(0);
	time_t t1 = start;

	// reading list mode
	while (true) {
		// reading next block
		int32_t size = reader.read();

		// exiting if no data
		if (size == 0) {
			break;
		}

		// processing block
		algorithm.process(reader.buffer(), size, events);
	}

	algorithm.finish(events);

	time_t t2 = time(0);

	cout << "terminated in " << (t2 - t1) << "s" << endl;

	cout << "#events read            = " << algorithm.processor().numberOfGoodEvents() << " (" << ((float)algorithm.processor().numberOfGoodEvents() / (float)(t2 - t1)) << " event/s)" << endl;
	cout << "#events bad             = " << algorithm.processor().numberOfBadEvents() << " (" << percentage(algorithm.processor().numberOfBadEvents(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;
	cout << "#events not inserted    = " << algorithm.processor().numberOfEventsNotInserted() << " (" << percentage(algorithm.processor().numberOfEventsNotInserted(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;
	cout << "#coincidences truncated = " << algorithm.processor().numberOfTruncatedCoincidences() << endl;
	cout << endl;

	algorithm.processor().printRates();

	cout << endl;
	cout << "max order = " << algorithm.processor().maxOrder() << endl;
	vector<Event> eventCoincidenceArray = algorithm.processor().maxEventCoincidenceArray();
	cout << "max coincidence" << endl;
	for (vector<Event>::const_iterator e = eventCoincidenceArray.begin(); e != eventCoincidenceArray.end(); ++e) {
		cout << *e << " = " << algorithm.processor().detector(*e) << endl;
	}

	return EXIT_SUCCESS;
}
