/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://www.osor.eu/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LST2ROOT_H
#define LST2ROOT_H

#include <cstdlib>
#include <fstream>
#include <sstream>
#include <cstring>
#include <algorithm>
#include <time.h>
#include <iomanip>

#include <TTree.h>
#include <TFile.h>

#include "../../Event.h"
#include "../../EventBlockArray.h"
#include "../../algorithms/root/Detector.h"
#include "../../Reader.h"
#include "../../EventArrayAlgorithm.h"
#include "../../algorithms/root/CoincidenceOffline.h"

#endif

using namespace std;
using namespace lstdpp128;

bool processFile(string inputFileName, time_type coincidenceDT, time_type cleanDT, int mult, string detectorADCFileName, string XPConfigFileName, int32_t blockSize, string outputDirectory, bool isVerbose, bool isSingle);

int32_t BUFFER_SIZE = 1 << 12;

int percentage(int a, int b) {
	return 100.0f * (float)a / (float)b;
}

void printRate(double * rate) {

	for (int i = 0; i < 7; ++i) {
		cout << rate[i] << " ";
	}
	cout << endl;
}

int main(int argc, char * argv[]) {

	if (argc < 6) {
		cerr << "usage : <file> <coincidence dT> <cleanDT> <mult> <detectoradc.txt> <XPConfig.txt> [blockSize=8192] [outputDirectory=.] [isVerbose=false] [isSingle=false]" << endl;
		return EXIT_FAILURE;
	}

	cout<<endl<<endl<<endl;
	cout<<"##################################################"<<endl;
	cout<<"#                 LST2ROOT                       #"<<endl;
	cout<<"##################################################"<<endl;

	string inputFile = argv[1];

	istringstream cit(argv[2]);
	time_type coincidenceDT;
	cit >> coincidenceDT;

	istringstream clit(argv[3]);
	time_type cleanDT;
	clit >> cleanDT;

	istringstream cmit(argv[4]);
	int mult;
	cmit >> mult;

	string detectorADCFileName = argv[5];

	string XPConfigFileName = argv[6];

	int32_t blockSize = BUFFER_SIZE * 2;

	if (argc >= 9) {
		istringstream im(argv[7]);
		im >> blockSize;
	}

	string outputDirectory = ".";
	if (argc >= 10) {
		outputDirectory = argv[8];
	}
	if (outputDirectory[outputDirectory.size()-1]!='/') outputDirectory+="/";

	bool isVerbose=false;
	if (argc >= 11) {
		string temp=argv[9];
		if (temp=="true"){
			isVerbose=true;
		}else{
			if (temp=="false"){
				isVerbose=false;
			}
		}
	}

	bool isSingle=false;
	if (argc >= 11) {
		string temp=argv[10];
		if (temp=="true"){
			isSingle=true;
		}else{
			if (temp=="false"){
				isSingle=false;
			}
		}
	}

	string inputFileName;
	string fileType=inputFile.substr(inputFile.find_last_of('.'), inputFile.size());
	if (fileType.substr(0,4)==".lst"){
		inputFileName=inputFile;
		bool success=processFile(inputFileName, coincidenceDT, cleanDT, mult, detectorADCFileName, XPConfigFileName, blockSize, outputDirectory, isVerbose, isSingle);
		if (!success){
			cout<<"ERROR: file '"<<inputFileName<<"' not processed"<<endl;
			cout<<"__________________________________________________"<<endl<<endl;
		}
	}else{
		if(fileType==".runlist"){
			ifstream file(inputFile.c_str());
			if (!file){
				cout<<"ERROR: run list '"<<inputFile<<"' does not exist"<<endl;
			}
			string line;
			while (getline(file, line)) {
				string testloop=line.substr(0,4);
				if (testloop=="loop"){
					if ((line.find(".lst") != line.npos)&&(line.rfind(".lst") != line.npos)){
						int firstrun, lastrun;
						istringstream issfirst(line.substr(line.find(".lst")-6,6));
						istringstream isslast(line.substr(line.rfind(".lst")-6,6));
						issfirst>>firstrun;
						isslast>>lastrun;

						for (int iRun=firstrun; iRun<=lastrun; iRun++){
							ostringstream convert;
							convert << setfill ('0') << setw (6) << iRun;
							string currentRun = convert.str()+".lst";
							bool success=processFile(currentRun, coincidenceDT, cleanDT, mult, detectorADCFileName, XPConfigFileName, blockSize, outputDirectory, isVerbose, isSingle);
							if (!success){
								cout<<"ERROR: file '"<<currentRun<<"' not processed"<<endl;
								cout<<"__________________________________________________"<<endl<<endl;
							}
						}


					}else{
						cout<<"ERROR: wrong loop command"<<endl;
					}
				}else{
					bool success=processFile(line, coincidenceDT, cleanDT, mult, detectorADCFileName, XPConfigFileName, blockSize, outputDirectory, isVerbose, isSingle);
					if (!success){
						cout<<"ERROR: file '"<<line<<"' not processed"<<endl;
						cout<<"__________________________________________________"<<endl<<endl;
					}
				}
			}
		}else{
			cerr << "usage : <file> should be .lst file or a .runlist" << endl;
			return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}

bool processFile(string inputFileName, time_type coincidenceDT, time_type cleanDT, int mult, string detectorADCFileName, string XPConfigFileName, int32_t blockSize, string outputDirectory, bool isVerbose, bool isSingle){

	cout<<endl<<"__________________________________________________"<<endl;
	cout<<"PROCESSING FILE '"<<inputFileName<<"'"<<endl<<endl;

	string fileType=inputFileName.substr(inputFileName.find_last_of('.'), 4);
	if (fileType!=".lst"){
		cerr << "usage : files in .runlist should be .lst files" << endl;
		return EXIT_FAILURE;
	}

	vector<int16_t> detectorADCList = loadList<int16_t>(detectorADCFileName);

	if (isVerbose){
		cout << "reading " << inputFileName << ", coincidence dT " << coincidenceDT << ", clean dT " << cleanDT << ", process block size " << blockSize << endl;
		cout << "detector ADC = " << printList<int16_t>(detectorADCList) << endl;
	}

	size_t pos = inputFileName.rfind('/');
	if (pos == string::npos){
		pos = 0;
	}else{
		pos++;
	}
	string outFileName =outputDirectory + inputFileName.substr(pos, inputFileName.find_last_of('.') - pos);


	Reader reader(BUFFER_SIZE);

	cout<<endl<<"> Opening file..."<<endl;
	bool success = reader.open(inputFileName);

	if (!success) {
		return false;
	}

	if (isVerbose){
		cout << listModeContext << endl;
	}

	// algorithm
	EventArrayAlgorithm<CoincidenceOffline> algorithm;
	algorithm.processor().init(detectorADCList, coincidenceDT, cleanDT, mult, blockSize, outFileName, isSingle); //AJOUT LST2ROOT

	// checking sizes
	if (algorithm.processor().getTotalNumberOfChannels() != detectorADCList.size()) {
		cout << "incoherent number of ADCs in the detector adc file and number of ADCs" << endl;
		cout << algorithm.processor().getTotalNumberOfChannels() << " adcs are expected from the data"<<endl;
		return false;
	}

	int nbADCsConfig=algorithm.processor().readConfigFile(XPConfigFileName);
	if (algorithm.processor().getTotalNumberOfChannels() != nbADCsConfig) {
		cout << "incoherent number of ADCs in the XPConfig file and number of ADCs" << endl;
		cout << algorithm.processor().getTotalNumberOfChannels() << " adcs are expected from the data"<<endl;
		return false;
	}

	// creating the events
	EventArray events(blockSize);

	int currentPercentage = 0;

	time_t start = time(0);
	time_t t1 = start;

	int32_t numberOfEvents = listModeContext.nbEvents;

	cout<<endl<<"> Reading and sorting events..."<<endl;
	// reading list mode
	while (true) {
		// reading next block
		int32_t size = reader.read();

		// exiting if no data
		if (size == 0) {
			break;
		}

		// processing block
		algorithm.process(reader.buffer(), size, events);
	}

	algorithm.finish(events);

	time_t t2 = time(0);
	if (isVerbose){
		cout << "terminated in " << (t2 - t1) << "s" << endl;

		cout << "#events read            = " << algorithm.processor().numberOfGoodEvents() << " (" << ((float)algorithm.processor().numberOfGoodEvents() / (float)(t2 - t1)) << " event/s)" << endl;
		cout << "#events not inserted    = " << algorithm.processor().numberOfEventsNotInserted() << " (" << percentage(algorithm.processor().numberOfEventsNotInserted(), algorithm.processor().numberOfGoodEvents()) << "%)" << endl;
		cout << "#coincidences truncated = " << algorithm.processor().numberOfTruncatedCoincidences() << endl;
		cout << endl;

		algorithm.processor().printRates();
	}

	cout<<"__________________________________________________"<<endl<<endl;

	return true;
}
